# encoding: utf-8
#
"""
License:

This file is part of MXCuBE.

MXCuBE is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MXCuBE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with MXCuBE. If not, see <https://www.gnu.org/licenses/>.
"""

__copyright__ = """ Copyright © 2016 -  2023 MXCuBE Collaboration."""
__license__ = "LGPLv3+"
__author__ = "rhfogh"
__date__ = "30/04/2023"

import operator
from collections import (
    OrderedDict,
    namedtuple,
)

CrystalClassInfo = namedtuple(
    "CrystalClassInfo",
    (
        "number",
        "name",
        "bravais_lattice",
        "point_group",
        "crystal_system",
        "laue_group",
    ),
)


# Crystal families (one-letter codes) compatible with a solution in a given family
# These are the solutions that could apply using the same axes,
# hence "c" not compatible with "h
SUB_LATTICE_MAP = {
    "a": "a",
    "m": "am",
    "o": "amo",
    "t": "amot",
    "h": "amh",
    "c": "amotc",
}

# Data from https://onlinelibrary.wiley.com/iucr/itc/Cb/ch1o4v0001/
# Laue group names from http://pd.chem.ucl.ac.uk/pdnn/symm2/laue1.htm
# Point group names from https://en.wikipedia.org/wiki/Crystallographic_point_group
CRYSTAL_CLASS_DATA = [
    CrystalClassInfo(0, "", "", "", "", ""),  # Null member, so number == index
    CrystalClassInfo(1, "1P", "aP", "1", "Triclinic", "-1"),
    CrystalClassInfo(2, "-1P", "aP", "-1", "Triclinic", "-1"),
    CrystalClassInfo(3, "2P", "mP", "2", "Monoclinic", "2/m"),
    CrystalClassInfo(4, "2C", "mC", "2", "Monoclinic", "2/m"),
    CrystalClassInfo(5, "mP", "mP", "m", "Monoclinic", "2/m"),
    CrystalClassInfo(6, "mC", "mC", "m", "Monoclinic", "2/m"),
    CrystalClassInfo(7, "2/mP", "mP", "2/m", "Monoclinic", "2/m"),
    CrystalClassInfo(8, "2/mC", "mC", "2/m", "Monoclinic", "2/m"),
    CrystalClassInfo(9, "222P", "oP", "222", "Orthorhombic", "mmm"),
    CrystalClassInfo(10, "222C", "oC", "222", "Orthorhombic", "mmm"),
    CrystalClassInfo(11, "222F", "oF", "222", "Orthorhombic", "mmm"),
    CrystalClassInfo(12, "222I", "oI", "222", "Orthorhombic", "mmm"),
    CrystalClassInfo(13, "mm2P", "oP", "mm2", "Orthorhombic", "mmm"),
    CrystalClassInfo(14, "mm2C", "oC", "mm2", "Orthorhombic", "mmm"),
    CrystalClassInfo(
        15, "2mmC", "oC", "mm2", "Orthorhombic", "mmm"
    ),  # NB was2mmC(Amm2)
    CrystalClassInfo(16, "mm2F", "oF", "mm2", "Orthorhombic", "mmm"),
    CrystalClassInfo(17, "mm2I", "oI", "mm2", "Orthorhombic", "mmm"),
    CrystalClassInfo(18, "mmmP", "oP", "mmm", "Orthorhombic", "mmm"),
    CrystalClassInfo(19, "mmmC", "oC", "mmm", "Orthorhombic", "mmm"),
    CrystalClassInfo(20, "mmmF", "oF", "mmm", "Orthorhombic", "mmm"),
    CrystalClassInfo(21, "mmmI", "oI", "mmm", "Orthorhombic", "mmm"),
    CrystalClassInfo(22, "4P", "tP", "4", "Tetragonal", "4/m"),
    CrystalClassInfo(23, "4I", "tI", "4", "Tetragonal", "4/m"),
    CrystalClassInfo(24, "-4P", "tP", "-4", "Tetragonal", "4/m"),
    CrystalClassInfo(25, "-4I", "tI", "-4", "Tetragonal", "4/m"),
    CrystalClassInfo(26, "4/mP", "tP", "4/m", "Tetragonal", "4/m"),
    CrystalClassInfo(27, "4/mI", "tI", "4/m", "Tetragonal", "4/m"),
    CrystalClassInfo(28, "422P", "tP", "422", "Tetragonal", "4/mmm"),
    CrystalClassInfo(29, "422I", "tI", "422", "Tetragonal", "4/mmm"),
    CrystalClassInfo(30, "4mmP", "tP", "4mm", "Tetragonal", "4/mmm"),
    CrystalClassInfo(31, "4mmI", "tI", "4mm", "Tetragonal", "4/mmm"),
    CrystalClassInfo(32, "-42mP", "tP", "-42m", "Tetragonal", "4/mmm"),
    CrystalClassInfo(33, "-4m2P", "tP", "-42m", "Tetragonal", "4/mmm"),
    CrystalClassInfo(34, "-4m2I", "tI", "-42m", "Tetragonal", "4/mmm"),
    CrystalClassInfo(35, "-42mI", "tI", "-42m", "Tetragonal", "4/mmm"),
    CrystalClassInfo(36, "4/mmmP", "tP", "4/mmm", "Tetragonal", "4/mmm"),
    CrystalClassInfo(37, "4/mmmI", "tI", "4/mmm", "Tetragonal", "4/mmm"),
    CrystalClassInfo(38, "3P", "hP", "3", "Trigonal", "-3"),
    CrystalClassInfo(39, "3R", "hR", "3", "Trigonal", "-3"),
    CrystalClassInfo(40, "-3P", "hP", "-3", "Trigonal", "-3"),
    CrystalClassInfo(41, "-3R", "hR", "-3", "Trigonal", "-3"),
    CrystalClassInfo(42, "312P", "hP", "32", "Trigonal", "-3m"),
    CrystalClassInfo(43, "321P", "hP", "32", "Trigonal", "-3m"),
    CrystalClassInfo(44, "32R", "hR", "32", "Trigonal", "-3m"),
    CrystalClassInfo(45, "3m1P", "hP", "3m", "Trigonal", "-3m"),
    CrystalClassInfo(46, "31mP", "hP", "3m", "Trigonal", "-3m"),
    CrystalClassInfo(47, "3mR", "hR", "3m", "Trigonal", "-3m"),
    CrystalClassInfo(48, "-31mP", "hP", "-3m", "Trigonal", "-3m"),
    CrystalClassInfo(49, "-3m1P", "hP", "-3m", "Trigonal", "-3m"),
    CrystalClassInfo(50, "-3mR", "hR", "-3m", "Trigonal", "-3m"),
    CrystalClassInfo(51, "6P", "hP", "6", "Hexagonal", "6/m"),
    CrystalClassInfo(52, "-6P", "hP", "-6", "Hexagonal", "6/m"),
    CrystalClassInfo(53, "6/mP", "hP", "6/m", "Hexagonal", "6/m"),
    CrystalClassInfo(54, "622P", "hP", "622", "Hexagonal", "6/mmm"),
    CrystalClassInfo(55, "6mmP", "hP", "6mm", "Hexagonal", "6/mmm"),
    CrystalClassInfo(56, "-6m2P", "hP", "-6m2", "Hexagonal", "6/mmm"),
    CrystalClassInfo(57, "-62mP", "hP", "-6m2", "Hexagonal", "6/mmm"),
    CrystalClassInfo(58, "6/mmmP", "hP", "6/mmm", "Hexagonal", "6/mmm"),
    CrystalClassInfo(59, "23P", "cP", "23", "Cubic", "m-3"),
    CrystalClassInfo(60, "23F", "cF", "23", "Cubic", "m-3"),
    CrystalClassInfo(61, "23I", "cI", "23", "Cubic", "m-3"),
    CrystalClassInfo(62, "m-3P", "cP", "m-3", "Cubic", "m-3"),
    CrystalClassInfo(63, "m-3F", "cF", "m-3", "Cubic", "m-3"),
    CrystalClassInfo(64, "m-3I", "cI", "m-3", "Cubic", "m-3"),
    CrystalClassInfo(65, "432P", "cP", "432", "Cubic", "m-3m"),
    CrystalClassInfo(66, "432F", "cF", "432", "Cubic", "m-3m"),
    CrystalClassInfo(67, "432I", "cI", "432", "Cubic", "m-3m"),
    CrystalClassInfo(68, "-43mP", "cP", "-43m", "Cubic", "m-3m"),
    CrystalClassInfo(69, "-43mF", "cF", "-43m", "Cubic", "m-3m"),
    CrystalClassInfo(70, "-43mI", "cI", "-43m", "Cubic", "m-3m"),
    CrystalClassInfo(71, "m-3mP", "cP", "m-3m", "Cubic", "m-3m"),
    CrystalClassInfo(72, "m-3mF", "cF", "m-3m", "Cubic", "m-3m"),
    CrystalClassInfo(73, "m-3mI", "cI", "m-3m", "Cubic", "m-3m"),
]
CRYSTAL_CLASS_MAP = OrderedDict((info.name, info) for info in CRYSTAL_CLASS_DATA)

SpaceGroupInfo = namedtuple(
    "SpaceGroupInfo", ("number", "name", "crystal_class", "synonyms")
)
# Names and synonyms derived from CCP4 (v9.) symop.lib
# Additional synonyms added for SG3-15 (e.g. 'P 21')
# Alternative axis orders added as synonyms for SG17,18
# SG 146,148,155,160,161,166,167 are given names starting with 'R',
# with 'H' names as synonyms
# Crystal class names from https://onlinelibrary.wiley.com/iucr/itc/Cb/ch1o4v0001/
SPACEGROUP_DATA = [
    SpaceGroupInfo(number=1, name="P1", crystal_class="1P", synonyms=("P 1",)),
    SpaceGroupInfo(number=2, name="P-1", crystal_class="-1P", synonyms=("P -1",)),
    SpaceGroupInfo(
        number=3, name="P2", crystal_class="2P", synonyms=("P 1 2 1", "P121", "P 2")
    ),
    SpaceGroupInfo(
        number=4, name="P21", crystal_class="2P", synonyms=("P 1 21 1", "P1211", "P 21")
    ),
    SpaceGroupInfo(
        number=5, name="C2", crystal_class="2C", synonyms=("C 1 2 1", "C121", "C 2")
    ),
    SpaceGroupInfo(
        number=6, name="Pm", crystal_class="mP", synonyms=("P 1 m 1", "P1m1", "P m")
    ),
    SpaceGroupInfo(
        number=7, name="Pc", crystal_class="mP", synonyms=("P 1 c 1", "P1c1", "P c")
    ),
    SpaceGroupInfo(
        number=8, name="Cm", crystal_class="mC", synonyms=("C 1 m 1", "C1m1", "C m")
    ),
    SpaceGroupInfo(
        number=9, name="Cc", crystal_class="mC", synonyms=("C 1 c 1", "C1c1", "C c")
    ),
    SpaceGroupInfo(
        number=10,
        name="P2/m",
        crystal_class="2/mP",
        synonyms=("P 1 2/m 1", "P12/m1", "P 2/m"),
    ),
    SpaceGroupInfo(
        number=11,
        name="P21/m",
        crystal_class="2/mP",
        synonyms=("P 1 21/m 1", "P121/m1", "P 21/m"),
    ),
    SpaceGroupInfo(
        number=12,
        name="C2/m",
        crystal_class="2/mC",
        synonyms=("C 1 2/m 1", "C12/m1", "C 2/m"),
    ),
    SpaceGroupInfo(
        number=13,
        name="P2/c",
        crystal_class="2/mP",
        synonyms=("P 1 2/c 1", "P12/c1", "P 2/c"),
    ),
    SpaceGroupInfo(
        number=14,
        name="P21/c",
        crystal_class="2/mP",
        synonyms=("P 1 21/c 1", "P121/c1", "P 21/c"),
    ),
    SpaceGroupInfo(
        number=15,
        name="C2/c",
        crystal_class="2/mC",
        synonyms=("C 1 2/c 1", "C12/c1", "C 2/c"),
    ),
    SpaceGroupInfo(number=16, name="P222", crystal_class="222P", synonyms=("P 2 2 2",)),
    SpaceGroupInfo(
        number=17,
        name="P2221",
        crystal_class="222P",
        synonyms=("P 2 2 21", "P2212", "P 2 21 2", "P2122", "P 21 2 2"),
    ),
    SpaceGroupInfo(
        number=18,
        name="P21212",
        crystal_class="222P",
        synonyms=("P 21 21 2", "P21221", "P 21 2 21", "P22121", "P 2 21 21"),
    ),
    SpaceGroupInfo(
        number=19, name="P212121", crystal_class="222P", synonyms=("P 21 21 21",)
    ),
    SpaceGroupInfo(
        number=20, name="C2221", crystal_class="222C", synonyms=("C 2 2 21",)
    ),
    SpaceGroupInfo(number=21, name="C222", crystal_class="222C", synonyms=("C 2 2 2",)),
    SpaceGroupInfo(number=22, name="F222", crystal_class="222F", synonyms=("F 2 2 2",)),
    SpaceGroupInfo(number=23, name="I222", crystal_class="222I", synonyms=("I 2 2 2",)),
    SpaceGroupInfo(
        number=24, name="I212121", crystal_class="222I", synonyms=("I 21 21 21",)
    ),
    SpaceGroupInfo(number=25, name="Pmm2", crystal_class="mm2P", synonyms=("P m m 2",)),
    SpaceGroupInfo(
        number=26, name="Pmc21", crystal_class="mm2P", synonyms=("P m c 21",)
    ),
    SpaceGroupInfo(number=27, name="Pcc2", crystal_class="mm2P", synonyms=("P c c 2",)),
    SpaceGroupInfo(number=28, name="Pma2", crystal_class="mm2P", synonyms=("P m a 2",)),
    SpaceGroupInfo(
        number=29, name="Pca21", crystal_class="mm2P", synonyms=("P c a 21",)
    ),
    SpaceGroupInfo(number=30, name="Pnc2", crystal_class="mm2P", synonyms=("P n c 2",)),
    SpaceGroupInfo(
        number=31, name="Pmn21", crystal_class="mm2P", synonyms=("P m n 21",)
    ),
    SpaceGroupInfo(number=32, name="Pba2", crystal_class="mm2P", synonyms=("P b a 2",)),
    SpaceGroupInfo(
        number=33, name="Pna21", crystal_class="mm2P", synonyms=("P n a 21",)
    ),
    SpaceGroupInfo(number=34, name="Pnn2", crystal_class="mm2P", synonyms=("P n n 2",)),
    SpaceGroupInfo(number=35, name="Cmm2", crystal_class="mm2C", synonyms=("C m m 2",)),
    SpaceGroupInfo(
        number=36, name="Cmc21", crystal_class="mm2C", synonyms=("C m c 21",)
    ),
    SpaceGroupInfo(number=37, name="Ccc2", crystal_class="mm2C", synonyms=("C c c 2",)),
    SpaceGroupInfo(number=38, name="Amm2", crystal_class="2mmC", synonyms=("A m m 2",)),
    SpaceGroupInfo(number=39, name="Abm2", crystal_class="2mmC", synonyms=("A b m 2",)),
    SpaceGroupInfo(number=40, name="Ama2", crystal_class="2mmC", synonyms=("A m a 2",)),
    SpaceGroupInfo(number=41, name="Aba2", crystal_class="2mmC", synonyms=("A b a 2",)),
    SpaceGroupInfo(number=42, name="Fmm2", crystal_class="mm2F", synonyms=("F m m 2",)),
    SpaceGroupInfo(number=43, name="Fdd2", crystal_class="mm2F", synonyms=("F d d 2",)),
    SpaceGroupInfo(number=44, name="Imm2", crystal_class="mm2I", synonyms=("I m m 2",)),
    SpaceGroupInfo(number=45, name="Iba2", crystal_class="mm2I", synonyms=("I b a 2",)),
    SpaceGroupInfo(number=46, name="Ima2", crystal_class="mm2I", synonyms=("I m a 2",)),
    SpaceGroupInfo(
        number=47,
        name="Pmmm",
        crystal_class="mmmP",
        synonyms=("P 2/m 2/m 2/m", "P2/m2/m2/m", "P m m m"),
    ),
    SpaceGroupInfo(
        number=48,
        name="Pnnn",
        crystal_class="mmmP",
        synonyms=("P 2/n 2/n 2/n", "P2/n2/n2/n", "P n n n"),
    ),
    SpaceGroupInfo(
        number=49,
        name="Pccm",
        crystal_class="mmmP",
        synonyms=("P 2/c 2/c 2/m", "P2/c2/c2/m", "P c c m"),
    ),
    SpaceGroupInfo(
        number=50,
        name="Pban",
        crystal_class="mmmP",
        synonyms=("P 2/b 2/a 2/n", "P2/b2/a2/n", "P b a n"),
    ),
    SpaceGroupInfo(
        number=51,
        name="Pmma",
        crystal_class="mmmP",
        synonyms=("P 21/m 2/m 2/a", "P21/m2/m2/a", "P m m a"),
    ),
    SpaceGroupInfo(
        number=52,
        name="Pnna",
        crystal_class="mmmP",
        synonyms=("P 2/n 21/n 2/a", "P2/n21/n2/a", "P n n a"),
    ),
    SpaceGroupInfo(
        number=53,
        name="Pmna",
        crystal_class="mmmP",
        synonyms=("P 2/m 2/n 21/a", "P2/m2/n21/a", "P m n a"),
    ),
    SpaceGroupInfo(
        number=54,
        name="Pcca",
        crystal_class="mmmP",
        synonyms=("P 21/c 2/c 2/a", "P21/c2/c2/a", "P c c a"),
    ),
    SpaceGroupInfo(
        number=55,
        name="Pbam",
        crystal_class="mmmP",
        synonyms=("P 21/b 21/a 2/m", "P21/b21/a2/m", "P b a m"),
    ),
    SpaceGroupInfo(
        number=56,
        name="Pccn",
        crystal_class="mmmP",
        synonyms=("P 21/c 21/c 2/n", "P21/c21/c2/n", "P c c n"),
    ),
    SpaceGroupInfo(
        number=57,
        name="Pbcm",
        crystal_class="mmmP",
        synonyms=("P 2/b 21/c 21/m", "P2/b21/c21/m", "P b c m"),
    ),
    SpaceGroupInfo(
        number=58,
        name="Pnnm",
        crystal_class="mmmP",
        synonyms=("P 21/n 21/n 2/m", "P21/n21/n2/m", "P n n m"),
    ),
    SpaceGroupInfo(
        number=59,
        name="Pmmn",
        crystal_class="mmmP",
        synonyms=("P 21/m 21/m 2/n", "P21/m21/m2/n", "P m m n"),
    ),
    SpaceGroupInfo(
        number=60,
        name="Pbcn",
        crystal_class="mmmP",
        synonyms=("P 21/b 2/c 21/n", "P21/b2/c21/n", "P b c n"),
    ),
    SpaceGroupInfo(
        number=61,
        name="Pbca",
        crystal_class="mmmP",
        synonyms=("P 21/b 21/c 21/a", "P21/b21/c21/a", "P b c a"),
    ),
    SpaceGroupInfo(
        number=62,
        name="Pnma",
        crystal_class="mmmP",
        synonyms=("P 21/n 21/m 21/a", "P21/n21/m21/a", "P n m a"),
    ),
    SpaceGroupInfo(
        number=63,
        name="Cmcm",
        crystal_class="mmmC",
        synonyms=("C 2/m 2/c 21/m", "C2/m2/c21/m", "C m c m"),
    ),
    SpaceGroupInfo(
        number=64,
        name="Cmca",
        crystal_class="mmmC",
        synonyms=("C 2/m 2/c 21/a", "C2/m2/c21/a", "C m c a"),
    ),
    SpaceGroupInfo(
        number=65,
        name="Cmmm",
        crystal_class="mmmC",
        synonyms=("C 2/m 2/m 2/m", "C2/m2/m2/m", "C m m m"),
    ),
    SpaceGroupInfo(
        number=66,
        name="Cccm",
        crystal_class="mmmC",
        synonyms=("C 2/c 2/c 2/m", "C2/c2/c2/m", "C c c m"),
    ),
    SpaceGroupInfo(
        number=67,
        name="Cmma",
        crystal_class="mmmC",
        synonyms=("C 2/m 2/m 2/a", "C2/m2/m2/a", "C m m a"),
    ),
    SpaceGroupInfo(
        number=68,
        name="Ccca",
        crystal_class="mmmC",
        synonyms=("C 2/c 2/c 2/a", "C2/c2/c2/a", "C c c a"),
    ),
    SpaceGroupInfo(
        number=69,
        name="Fmmm",
        crystal_class="mmmF",
        synonyms=("F 2/m 2/m 2/m", "F2/m2/m2/m", "F m m m"),
    ),
    SpaceGroupInfo(
        number=70,
        name="Fddd",
        crystal_class="mmmF",
        synonyms=("F 2/d 2/d 2/d", "F2/d2/d2/d", "F d d d"),
    ),
    SpaceGroupInfo(
        number=71,
        name="Immm",
        crystal_class="mmmI",
        synonyms=("I 2/m 2/m 2/m", "I2/m2/m2/m", "I m m m"),
    ),
    SpaceGroupInfo(
        number=72,
        name="Ibam",
        crystal_class="mmmI",
        synonyms=("I 2/b 2/a 2/m", "I2/b2/a2/m", "I b a m"),
    ),
    SpaceGroupInfo(
        number=73,
        name="Ibca",
        crystal_class="mmmI",
        synonyms=("I 21/b 21/c 21/a", "I21/b21/c21/a", "I b c a"),
    ),
    SpaceGroupInfo(
        number=74,
        name="Imma",
        crystal_class="mmmI",
        synonyms=("I 21/m 21/m 21/a", "I21/m21/m21/a", "I m m a"),
    ),
    SpaceGroupInfo(number=75, name="P4", crystal_class="4P", synonyms=("P 4",)),
    SpaceGroupInfo(number=76, name="P41", crystal_class="4P", synonyms=("P 41",)),
    SpaceGroupInfo(number=77, name="P42", crystal_class="4P", synonyms=("P 42",)),
    SpaceGroupInfo(number=78, name="P43", crystal_class="4P", synonyms=("P 43",)),
    SpaceGroupInfo(number=79, name="I4", crystal_class="4I", synonyms=("I 4",)),
    SpaceGroupInfo(number=80, name="I41", crystal_class="4I", synonyms=("I 41",)),
    SpaceGroupInfo(number=81, name="P-4", crystal_class="-4P", synonyms=("P -4",)),
    SpaceGroupInfo(number=82, name="I-4", crystal_class="-4I", synonyms=("I -4",)),
    SpaceGroupInfo(number=83, name="P4/m", crystal_class="4/mP", synonyms=("P 4/m",)),
    SpaceGroupInfo(number=84, name="P42/m", crystal_class="4/mP", synonyms=("P 42/m",)),
    SpaceGroupInfo(number=85, name="P4/n", crystal_class="4/mP", synonyms=("P 4/n",)),
    SpaceGroupInfo(number=86, name="P42/n", crystal_class="4/mP", synonyms=("P 42/n",)),
    SpaceGroupInfo(number=87, name="I4/m", crystal_class="4/mI", synonyms=("I 4/m",)),
    SpaceGroupInfo(number=88, name="I41/a", crystal_class="4/mI", synonyms=("I 41/a",)),
    SpaceGroupInfo(number=89, name="P422", crystal_class="422P", synonyms=("P 4 2 2",)),
    SpaceGroupInfo(
        number=90, name="P4212", crystal_class="422P", synonyms=("P 4 21 2",)
    ),
    SpaceGroupInfo(
        number=91, name="P4122", crystal_class="422P", synonyms=("P 41 2 2",)
    ),
    SpaceGroupInfo(
        number=92, name="P41212", crystal_class="422P", synonyms=("P 41 21 2",)
    ),
    SpaceGroupInfo(
        number=93, name="P4222", crystal_class="422P", synonyms=("P 42 2 2",)
    ),
    SpaceGroupInfo(
        number=94, name="P42212", crystal_class="422P", synonyms=("P 42 21 2",)
    ),
    SpaceGroupInfo(
        number=95, name="P4322", crystal_class="422P", synonyms=("P 43 2 2",)
    ),
    SpaceGroupInfo(
        number=96, name="P43212", crystal_class="422P", synonyms=("P 43 21 2",)
    ),
    SpaceGroupInfo(number=97, name="I422", crystal_class="422I", synonyms=("I 4 2 2",)),
    SpaceGroupInfo(
        number=98, name="I4122", crystal_class="422I", synonyms=("I 41 2 2",)
    ),
    SpaceGroupInfo(number=99, name="P4mm", crystal_class="4mmP", synonyms=("P 4 m m",)),
    SpaceGroupInfo(
        number=100, name="P4bm", crystal_class="4mmP", synonyms=("P 4 b m",)
    ),
    SpaceGroupInfo(
        number=101, name="P42cm", crystal_class="4mmP", synonyms=("P 42 c m",)
    ),
    SpaceGroupInfo(
        number=102, name="P42nm", crystal_class="4mmP", synonyms=("P 42 n m",)
    ),
    SpaceGroupInfo(
        number=103, name="P4cc", crystal_class="4mmP", synonyms=("P 4 c c",)
    ),
    SpaceGroupInfo(
        number=104, name="P4nc", crystal_class="4mmP", synonyms=("P 4 n c",)
    ),
    SpaceGroupInfo(
        number=105, name="P42mc", crystal_class="4mmP", synonyms=("P 42 m c",)
    ),
    SpaceGroupInfo(
        number=106, name="P42bc", crystal_class="4mmP", synonyms=("P 42 b c",)
    ),
    SpaceGroupInfo(
        number=107, name="I4mm", crystal_class="4mmI", synonyms=("I 4 m m",)
    ),
    SpaceGroupInfo(
        number=108, name="I4cm", crystal_class="4mmI", synonyms=("I 4 c m",)
    ),
    SpaceGroupInfo(
        number=109, name="I41md", crystal_class="4mmI", synonyms=("I 41 m d",)
    ),
    SpaceGroupInfo(
        number=110, name="I41cd", crystal_class="4mmI", synonyms=("I 41 c d",)
    ),
    SpaceGroupInfo(
        number=111, name="P-42m", crystal_class="-42mP", synonyms=("P -4 2 m",)
    ),
    SpaceGroupInfo(
        number=112, name="P-42c", crystal_class="-42mP", synonyms=("P -4 2 c",)
    ),
    SpaceGroupInfo(
        number=113, name="P-421m", crystal_class="-42mP", synonyms=("P -4 21 m",)
    ),
    SpaceGroupInfo(
        number=114, name="P-421c", crystal_class="-42mP", synonyms=("P -4 21 c",)
    ),
    SpaceGroupInfo(
        number=115, name="P-4m2", crystal_class="-4m2P", synonyms=("P -4 m 2",)
    ),
    SpaceGroupInfo(
        number=116, name="P-4c2", crystal_class="-4m2P", synonyms=("P -4 c 2",)
    ),
    SpaceGroupInfo(
        number=117, name="P-4b2", crystal_class="-4m2P", synonyms=("P -4 b 2",)
    ),
    SpaceGroupInfo(
        number=118, name="P-4n2", crystal_class="-4m2P", synonyms=("P -4 n 2",)
    ),
    SpaceGroupInfo(
        number=119, name="I-4m2", crystal_class="-4m2I", synonyms=("I -4 m 2",)
    ),
    SpaceGroupInfo(
        number=120, name="I-4c2", crystal_class="-4m2I", synonyms=("I -4 c 2",)
    ),
    SpaceGroupInfo(
        number=121, name="I-42m", crystal_class="-42mI", synonyms=("I -4 2 m",)
    ),
    SpaceGroupInfo(
        number=122, name="I-42d", crystal_class="-42mI", synonyms=("I -4 2 d",)
    ),
    SpaceGroupInfo(
        number=123,
        name="P4/mmm",
        crystal_class="4/mmmP",
        synonyms=("P 4/m 2/m 2/m", "P4/m2/m2/m", "P4/m m m"),
    ),
    SpaceGroupInfo(
        number=124,
        name="P4/mcc",
        crystal_class="4/mmmP",
        synonyms=("P 4/m 2/c 2/c", "P4/m2/c2/c", "P4/m c c"),
    ),
    SpaceGroupInfo(
        number=125,
        name="P4/nbm",
        crystal_class="4/mmmP",
        synonyms=("P 4/n 2/b 2/m", "P4/n2/b2/m", "P4/n b m"),
    ),
    SpaceGroupInfo(
        number=126,
        name="P4/nnc",
        crystal_class="4/mmmP",
        synonyms=("P 4/n 2/n 2/c", "P4/n2/n2/c", "P4/n n c"),
    ),
    SpaceGroupInfo(
        number=127,
        name="P4/mbm",
        crystal_class="4/mmmP",
        synonyms=("P 4/m 21/b 2/m", "P4/m21/b2/m", "P4/m b m"),
    ),
    SpaceGroupInfo(
        number=128,
        name="P4/mnc",
        crystal_class="4/mmmP",
        synonyms=("P 4/m 21/n 2/c", "P4/m21/n2/c", "P4/m n c"),
    ),
    SpaceGroupInfo(
        number=129,
        name="P4/nmm",
        crystal_class="4/mmmP",
        synonyms=("P 4/n 21/m 2/m", "P4/n21/m2/m", "P4/n m m"),
    ),
    SpaceGroupInfo(
        number=130,
        name="P4/ncc",
        crystal_class="4/mmmP",
        synonyms=("P 4/n 2/c 2/c", "P4/n2/c2/c", "P4/n c c"),
    ),
    SpaceGroupInfo(
        number=131,
        name="P42/mmc",
        crystal_class="4/mmmP",
        synonyms=("P 42/m 2/m 2/c", "P42/m2/m2/c", "P42/m m c"),
    ),
    SpaceGroupInfo(
        number=132,
        name="P42/mcm",
        crystal_class="4/mmmP",
        synonyms=("P 42/m 2/c 2/m", "P42/m2/c2/m", "P42/m c m"),
    ),
    SpaceGroupInfo(
        number=133,
        name="P42/nbc",
        crystal_class="4/mmmP",
        synonyms=("P 42/n 2/b 2/c", "P42/n2/b2/c", "P42/n b c"),
    ),
    SpaceGroupInfo(
        number=134,
        name="P42/nnm",
        crystal_class="4/mmmP",
        synonyms=("P 42/n 2/n 2/m", "P42/n2/n2/m", "P42/n n m"),
    ),
    SpaceGroupInfo(
        number=135,
        name="P42/mbc",
        crystal_class="4/mmmP",
        synonyms=("P 42/m 21/b 2/c", "P42/m21/b2/c", "P42/m b c"),
    ),
    SpaceGroupInfo(
        number=136,
        name="P42/mnm",
        crystal_class="4/mmmP",
        synonyms=("P 42/m 21/n 2/m", "P42/m21/n2/m", "P42/m n m"),
    ),
    SpaceGroupInfo(
        number=137,
        name="P42/nmc",
        crystal_class="4/mmmP",
        synonyms=("P 42/n 21/m 2/c", "P42/n21/m2/c", "P42/n m c"),
    ),
    SpaceGroupInfo(
        number=138,
        name="P42/ncm",
        crystal_class="4/mmmP",
        synonyms=("P 42/n 21/c 2/m", "P42/n21/c2/m", "P42/n c m"),
    ),
    SpaceGroupInfo(
        number=139,
        name="I4/mmm",
        crystal_class="4/mmmI",
        synonyms=("I 4/m 2/m 2/m", "I4/m2/m2/m", "I4/m m m"),
    ),
    SpaceGroupInfo(
        number=140,
        name="I4/mcm",
        crystal_class="4/mmmI",
        synonyms=("I 4/m 2/c 2/m", "I4/m2/c2/m", "I4/m c m"),
    ),
    SpaceGroupInfo(
        number=141,
        name="I41/amd",
        crystal_class="4/mmmI",
        synonyms=("I 41/a 2/m 2/d", "I41/a2/m2/d", "I41/a m d"),
    ),
    SpaceGroupInfo(
        number=142,
        name="I41/acd",
        crystal_class="4/mmmI",
        synonyms=("I 41/a 2/c 2/d", "I41/a2/c2/d", "I41/a c d"),
    ),
    SpaceGroupInfo(number=143, name="P3", crystal_class="3P", synonyms=("P 3",)),
    SpaceGroupInfo(number=144, name="P31", crystal_class="3P", synonyms=("P 31",)),
    SpaceGroupInfo(number=145, name="P32", crystal_class="3P", synonyms=("P 32",)),
    SpaceGroupInfo(
        number=146, name="R3", crystal_class="3R", synonyms=("H 3", "H3", " R 3")
    ),
    SpaceGroupInfo(number=147, name="P-3", crystal_class="-3P", synonyms=("P -3",)),
    SpaceGroupInfo(
        number=148, name="R-3", crystal_class="-3R", synonyms=("H -3", "H-3", "R -3")
    ),
    SpaceGroupInfo(
        number=149, name="P312", crystal_class="312P", synonyms=("P 3 1 2",)
    ),
    SpaceGroupInfo(
        number=150, name="P321", crystal_class="321P", synonyms=("P 3 2 1",)
    ),
    SpaceGroupInfo(
        number=151, name="P3112", crystal_class="312P", synonyms=("P 31 1 2",)
    ),
    SpaceGroupInfo(
        number=152, name="P3121", crystal_class="321P", synonyms=("P 31 2 1",)
    ),
    SpaceGroupInfo(
        number=153, name="P3212", crystal_class="312P", synonyms=("P 32 1 2",)
    ),
    SpaceGroupInfo(
        number=154, name="P3221", crystal_class="321P", synonyms=("P 32 2 1",)
    ),
    SpaceGroupInfo(
        number=155, name="R32", crystal_class="32R", synonyms=("H 3 2", "H32", "R 3 2")
    ),
    SpaceGroupInfo(
        number=156, name="P3m1", crystal_class="3m1P", synonyms=("P 3 m 1",)
    ),
    SpaceGroupInfo(
        number=157, name="P31m", crystal_class="31mP", synonyms=("P 3 1 m",)
    ),
    SpaceGroupInfo(
        number=158, name="P3c1", crystal_class="3m1P", synonyms=("P 3 c 1",)
    ),
    SpaceGroupInfo(
        number=159, name="P31c", crystal_class="31mP", synonyms=("P 3 1 c",)
    ),
    SpaceGroupInfo(
        number=160, name="R3m", crystal_class="3mR", synonyms=("H 3 m", "H3m", "R 3 m")
    ),
    SpaceGroupInfo(
        number=161, name="R3c", crystal_class="3mR", synonyms=("H 3 c", "H3c", "R 3 c")
    ),
    SpaceGroupInfo(
        number=162,
        name="P-31m",
        crystal_class="-31mP",
        synonyms=("P -3 1 2/m", "P-312/m", "P -3 1 m"),
    ),
    SpaceGroupInfo(
        number=163,
        name="P-31c",
        crystal_class="-31mP",
        synonyms=("P -3 1 2/c", "P-312/c", "P -3 1 c"),
    ),
    SpaceGroupInfo(
        number=164,
        name="P-3m1",
        crystal_class="-3m1P",
        synonyms=("P -3 2/m 1", "P-32/m1", "P -3 m 1"),
    ),
    SpaceGroupInfo(
        number=165,
        name="P-3c1",
        crystal_class="-3m1P",
        synonyms=("P -3 2/c 1", "P-32/c1", "P -3 c 1"),
    ),
    SpaceGroupInfo(
        number=166,
        name="R-3m",
        crystal_class="-3mR",
        synonyms=(
            "H-3m",
            "H -3 2/m",
            "H-32/m",
            "H -3 m",
            "R -3 2/m",
            "R-32/m",
            "R -3 m",
        ),
    ),
    SpaceGroupInfo(
        number=167,
        name="R-3c",
        crystal_class="-3mR",
        synonyms=(
            "H-3c",
            "H -3 2/c",
            "H-32/c",
            "H -3 c",
            "R -3 2/c",
            "R-32/c",
            "R -3 c",
        ),
    ),
    SpaceGroupInfo(number=168, name="P6", crystal_class="6P", synonyms=("P 6",)),
    SpaceGroupInfo(number=169, name="P61", crystal_class="6P", synonyms=("P 61",)),
    SpaceGroupInfo(number=170, name="P65", crystal_class="6P", synonyms=("P 65",)),
    SpaceGroupInfo(number=171, name="P62", crystal_class="6P", synonyms=("P 62",)),
    SpaceGroupInfo(number=172, name="P64", crystal_class="6P", synonyms=("P 64",)),
    SpaceGroupInfo(number=173, name="P63", crystal_class="6P", synonyms=("P 63",)),
    SpaceGroupInfo(number=174, name="P-6", crystal_class="-6P", synonyms=("P -6",)),
    SpaceGroupInfo(number=175, name="P6/m", crystal_class="6/mP", synonyms=("P 6/m",)),
    SpaceGroupInfo(
        number=176, name="P63/m", crystal_class="6/mP", synonyms=("P 63/m",)
    ),
    SpaceGroupInfo(
        number=177, name="P622", crystal_class="622P", synonyms=("P 6 2 2",)
    ),
    SpaceGroupInfo(
        number=178, name="P6122", crystal_class="622P", synonyms=("P 61 2 2",)
    ),
    SpaceGroupInfo(
        number=179, name="P6522", crystal_class="622P", synonyms=("P 65 2 2",)
    ),
    SpaceGroupInfo(
        number=180, name="P6222", crystal_class="622P", synonyms=("P 62 2 2",)
    ),
    SpaceGroupInfo(
        number=181, name="P6422", crystal_class="622P", synonyms=("P 64 2 2",)
    ),
    SpaceGroupInfo(
        number=182, name="P6322", crystal_class="622P", synonyms=("P 63 2 2",)
    ),
    SpaceGroupInfo(
        number=183, name="P6mm", crystal_class="6mmP", synonyms=("P 6 m m",)
    ),
    SpaceGroupInfo(
        number=184, name="P6cc", crystal_class="6mmP", synonyms=("P 6 c c",)
    ),
    SpaceGroupInfo(
        number=185, name="P63cm", crystal_class="6mmP", synonyms=("P 63 c m",)
    ),
    SpaceGroupInfo(
        number=186, name="P63mc", crystal_class="6mmP", synonyms=("P 63 m c",)
    ),
    SpaceGroupInfo(
        number=187, name="P-6m2", crystal_class="-6m2P", synonyms=("P -6 m 2",)
    ),
    SpaceGroupInfo(
        number=188, name="P-6c2", crystal_class="-6m2P", synonyms=("P -6 c 2",)
    ),
    SpaceGroupInfo(
        number=189, name="P-62m", crystal_class="-62mP", synonyms=("P -6 2 m",)
    ),
    SpaceGroupInfo(
        number=190, name="P-62c", crystal_class="-62mP", synonyms=("P -6 2 c",)
    ),
    SpaceGroupInfo(
        number=191,
        name="P6/mmm",
        crystal_class="6/mmmP",
        synonyms=("P 6/m 2/m 2/m", "P6/m2/m2/m", "P 6/m m m"),
    ),
    SpaceGroupInfo(
        number=192,
        name="P6/mcc",
        crystal_class="6/mmmP",
        synonyms=("P 6/m 2/c 2/c", "P6/m2/c2/c", "P 6/m c c"),
    ),
    SpaceGroupInfo(
        number=193,
        name="P63/mcm",
        crystal_class="6/mmmP",
        synonyms=("P 63/m 2/c 2/m", "P63/m2/c2/m", "P 63/m c m"),
    ),
    SpaceGroupInfo(
        number=194,
        name="P63/mmc",
        crystal_class="6/mmmP",
        synonyms=("P 63/m 2/m 2/c", "P63/m2/m2/c", "P 63/m m c"),
    ),
    SpaceGroupInfo(number=195, name="P23", crystal_class="23P", synonyms=("P 2 3",)),
    SpaceGroupInfo(number=196, name="F23", crystal_class="23F", synonyms=("F 2 3",)),
    SpaceGroupInfo(number=197, name="I23", crystal_class="23I", synonyms=("I 2 3",)),
    SpaceGroupInfo(number=198, name="P213", crystal_class="23P", synonyms=("P 21 3",)),
    SpaceGroupInfo(number=199, name="I213", crystal_class="23I", synonyms=("I 21 3",)),
    SpaceGroupInfo(
        number=200,
        name="Pm-3",
        crystal_class="m-3P",
        synonyms=("P 2/m -3", "P2/m-3", "P m -3"),
    ),
    SpaceGroupInfo(
        number=201,
        name="Pn-3",
        crystal_class="m-3P",
        synonyms=("P 2/n -3", "P2/n-3", "P n -3"),
    ),
    SpaceGroupInfo(
        number=202,
        name="Fm-3",
        crystal_class="m-3F",
        synonyms=("F 2/m -3", "F2/m-3", "F m -3"),
    ),
    SpaceGroupInfo(
        number=203,
        name="Fd-3",
        crystal_class="m-3F",
        synonyms=("F 2/d -3", "F2/d-3", "F d -3"),
    ),
    SpaceGroupInfo(
        number=204,
        name="Im-3",
        crystal_class="m-3I",
        synonyms=("I 2/m -3", "I2/m-3", "I m -3"),
    ),
    SpaceGroupInfo(
        number=205,
        name="Pa-3",
        crystal_class="m-3P",
        synonyms=("P 21/a -3", "P21/a-3", "P a -3"),
    ),
    SpaceGroupInfo(
        number=206,
        name="Ia-3",
        crystal_class="m-3I",
        synonyms=("I 21/a -3", "I21/a-3", "I a -3"),
    ),
    SpaceGroupInfo(
        number=207, name="P432", crystal_class="432P", synonyms=("P 4 3 2",)
    ),
    SpaceGroupInfo(
        number=208, name="P4232", crystal_class="432P", synonyms=("P 42 3 2",)
    ),
    SpaceGroupInfo(
        number=209, name="F432", crystal_class="432F", synonyms=("F 4 3 2",)
    ),
    SpaceGroupInfo(
        number=210, name="F4132", crystal_class="432F", synonyms=("F 41 3 2",)
    ),
    SpaceGroupInfo(
        number=211, name="I432", crystal_class="432I", synonyms=("I 4 3 2",)
    ),
    SpaceGroupInfo(
        number=212, name="P4332", crystal_class="432P", synonyms=("P 43 3 2",)
    ),
    SpaceGroupInfo(
        number=213, name="P4132", crystal_class="432P", synonyms=("P 41 3 2",)
    ),
    SpaceGroupInfo(
        number=214, name="I4132", crystal_class="432I", synonyms=("I 41 3 2",)
    ),
    SpaceGroupInfo(
        number=215, name="P-43m", crystal_class="-43mP", synonyms=("P -4 3 m",)
    ),
    SpaceGroupInfo(
        number=216, name="F-43m", crystal_class="-43mF", synonyms=("F -4 3 m",)
    ),
    SpaceGroupInfo(
        number=217, name="I-43m", crystal_class="-43mI", synonyms=("I -4 3 m",)
    ),
    SpaceGroupInfo(
        number=218, name="P-43n", crystal_class="-43mP", synonyms=("P -4 3 n",)
    ),
    SpaceGroupInfo(
        number=219, name="F-43c", crystal_class="-43mF", synonyms=("F -4 3 c",)
    ),
    SpaceGroupInfo(
        number=220, name="I-43d", crystal_class="-43mI", synonyms=("I -4 3 d",)
    ),
    SpaceGroupInfo(
        number=221,
        name="Pm-3m",
        crystal_class="m-3mP",
        synonyms=("P 4/m -3 2/m", "P4/m-32/m", "P m -3 m"),
    ),
    SpaceGroupInfo(
        number=222,
        name="Pn-3n",
        crystal_class="m-3mP",
        synonyms=("P 4/n -3 2/n", "P4/n-32/n", "P n -3 n"),
    ),
    SpaceGroupInfo(
        number=223,
        name="Pm-3n",
        crystal_class="m-3mP",
        synonyms=("P 42/m -3 2/n", "P42/m-32/n", "P m -3 n"),
    ),
    SpaceGroupInfo(
        number=224,
        name="Pn-3m",
        crystal_class="m-3mP",
        synonyms=("P 42/n -3 2/m", "P42/n-32/m", "P n -3 m"),
    ),
    SpaceGroupInfo(
        number=225,
        name="Fm-3m",
        crystal_class="m-3mF",
        synonyms=("F 4/m -3 2/m", "F4/m-32/m", "F m -3 m"),
    ),
    SpaceGroupInfo(
        number=226,
        name="Fm-3c",
        crystal_class="m-3mF",
        synonyms=("F 4/m -3 2/c", "F4/m-32/c", "F m -3 c"),
    ),
    SpaceGroupInfo(
        number=227,
        name="Fd-3m",
        crystal_class="m-3mF",
        synonyms=("F 41/d -3 2/m", "F41/d-32/m", "F d -3 m"),
    ),
    SpaceGroupInfo(
        number=228,
        name="Fd-3c",
        crystal_class="m-3mF",
        synonyms=("F 41/d -3 2/c", "F41/d-32/c", "F d -3 c"),
    ),
    SpaceGroupInfo(
        number=229,
        name="Im-3m",
        crystal_class="m-3mI",
        synonyms=("I 4/m -3 2/m", "I4/m-32/m", "I m -3 m"),
    ),
    SpaceGroupInfo(
        number=230,
        name="Ia-3d",
        crystal_class="m-3mI",
        synonyms=("I 41/a -3 2/d", "I41/a-32/d", "I a -3 d"),
    ),
]
SPACEGROUP_MAP = OrderedDict((info.name, info) for info in SPACEGROUP_DATA)
for tpl in SPACEGROUP_DATA:
    # Done this way so that first elements in map are on item per spacegroup
    for tag in tpl.synonyms:
        SPACEGROUP_MAP[tag] = tpl

# Space group names for space groups compatible with chiral molecules,
# i.e. that do not contain mirror planes or centres of symmetry,
# in order of space group number
XTAL_SPACEGROUPS = [""] + [
    info.name
    for info in SPACEGROUP_DATA
    if CRYSTAL_CLASS_MAP[info.crystal_class].point_group.isdigit()
]

BRAVAIS_LATTICES = (
    "aP",
    "mP",
    "mC",
    "oP",
    "oC",
    "oF",
    "oI",
    "tP",
    "tI",
    "hP",
    "hR",
    "cP",
    "cF",
    "cI",
)
UI_LATTICES = BRAVAIS_LATTICES + ("mI",)


def filter_crystal_classes(bravais_lattice, crystal_classes=()):
    """Filter crystal classes to select those compatible with selected Bravais lattice

    including sublattices

    Args:
        bravais_lattice: str
        crystal_classes: Sequence

    Returns:
        tuople

    """
    compatibles = SUB_LATTICE_MAP[bravais_lattice[0]]
    result = tuple(
        xcls
        for xcls in crystal_classes
        if CRYSTAL_CLASS_MAP[xcls].bravais_lattice[0] in compatibles
    )
    return result


def space_groups_from_params(lattices=(), point_groups=(), chiral_only=True):
    """list of names of space groups compatible with lattices and point groups
    Given in space group number order

    Args:
        lattices:
        point_groups:
        chiral_only:

    Returns:

    """
    if chiral_only:
        space_groups = XTAL_SPACEGROUPS[1:]
    else:
        space_groups = list(info.name for info in SPACEGROUP_DATA)
    if lattices or point_groups:
        sgs1 = []
        if lattices:
            converter = {
                "Triclinic": "a",
                "Monoclinic": "m",
                "Orthorhombic": "o",
                "Tetragonal": "t",
                "Trigonal": "h",
                "Hexagonal": "h",
                "Cubic": "c",
            }
            tsts = set(converter.get(tag, tag) for tag in lattices)
            if "mI" in tsts:
                # Special case. mI is supported in XDS and UI but is not official
                tsts.add("mC")
            for spg in space_groups:
                blattice = CRYSTAL_CLASS_MAP[
                    SPACEGROUP_MAP[spg].crystal_class
                ].bravais_lattice
                if any(blattice.startswith(tst) for tst in tsts):
                    sgs1.append(spg)

        sgs2 = []
        if point_groups:
            for spg in space_groups:
                info = SPACEGROUP_MAP[spg]
                ccinfo = CRYSTAL_CLASS_MAP[(info.crystal_class)]
                for pgp in point_groups:
                    if ccinfo.point_group == pgp or (
                        pgp in ("312", "321") and ccinfo.name[:3] == pgp
                    ):
                        sgs2.append(spg)
                        break
        if sgs1 and sgs2:
            tstset = set(sgs1)
            space_groups = list(spg for spg in sgs2 if spg in tstset)
        else:
            space_groups = sgs1 + sgs2
    return space_groups


def crystal_classes_from_params(
    lattices: tuple = (), point_groups: tuple = (), space_group: str = None
):
    """
    Get tuple of crystal class names compatible with input parameters,
    in crystal class number order
    Raises error for incompatible data.
    NB If lattices or point_groups are set will return relevant tuple
    even if space_group is also set

    Args:
        lattices: list of lattice strings (Bravais Lattice or 'Monoclinic' etc.
        point_groups: list(str) List of point group names (or '312', '321')
        space_group:  Space group name

    Returns: tuple(str) of crystal class names

    """
    if lattices or point_groups:
        space_groups = space_groups_from_params(
            lattices=lattices, point_groups=point_groups
        )
        if not space_groups or (space_group and space_group not in space_groups):
            result = ()
        else:
            infos = frozenset(
                CRYSTAL_CLASS_MAP[SPACEGROUP_MAP[space_group].crystal_class]
                for space_group in space_groups
            )
            result = tuple(
                info.name for info in sorted(infos, key=operator.attrgetter("number"))
            )
    elif space_group:
        result = (SPACEGROUP_MAP[space_group].crystal_class,)
    else:
        # Return empty list (nothing is set)
        result = ()
    return result


def strategy_laue_group(crystal_classes: tuple, phasing=False):
    """Get laue group and point-group-like to use for strategy calculation
    for a given set of crystal classes.

    Args:
        crystal_classes (tuple): Crystal class names
        phasing (bool): Is this for phasing (default is native)

    Returns: (str, str) laue_group, point_group_like

    """

    laue_group_map = {
        frozenset(("-1",)): ("-1", "1"),
        frozenset(("2/m",)): ("2/m", "2"),
        frozenset(("mmm",)): ("mmm", "222"),
        frozenset(("4/m",)): ("4/m", "4"),
        frozenset(("4/mmm",)): ("4/mmm", "422"),
        frozenset(("4/m", "4/mmm")): ("4/m", "4"),
        frozenset(("-3",)): ("-3", "3"),
        frozenset(("6/m",)): ("6/m", "6"),
        frozenset(("6/mmm",)): ("6/mmm", "622"),
        frozenset(("6/m", "6/mmm")): ("6/m", "6"),
        frozenset(("m-3",)): ("m-3", "23"),
        frozenset(("m-3m",)): ("m-3m", "432"),
        frozenset(("m-3", "m-3m")): ("m-3", "23"),
    }
    # NB Other combinations are dealt with separately

    laue_groups = set()
    lattices = set()
    for name in crystal_classes:
        info = CRYSTAL_CLASS_MAP[name]
        laue_groups.add(info.laue_group)
        lattices.add(info.bravais_lattice)
    laue_groups = frozenset(laue_groups)
    result = laue_group_map.get(laue_groups)
    if result is None:
        if lattices and lattices.issubset(set(("hP", "hR"))):
            if laue_groups == set(("-3m",)) and lattices == set(("hR",)):
                # NB Includes non-chiral crystal classes
                result = ("-3m", "32")
            elif len(crystal_classes) == 1 and crystal_classes[0] in ("312P", "321P"):
                # NB excludes non-chiral crystal classes
                result = ("-3m", crystal_classes[0][:-1])
            else:
                result = ("-3", "3")
        else:
            result = ("-1", "1")
    return result


def regularise_space_group(sgname: str):
    """Convert input (ISPyB) space group name to official space group name"""

    sginfo = SPACEGROUP_MAP.get(sgname)
    if sginfo:
        return sginfo.name
    return None



"""Class for cameras connected to framegrabbers run by Taco Device Servers"""

import atexit
import logging
import os
import signal
import subprocess
import time
from typing import (
    List,
    Tuple,
)

import gevent
import psutil
from PIL import Image

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR

MAX_TRIES = 3
SLOW_INTERVAL = 1000


class MDCameraMockup(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def _init(self):
        self._format = "MPEG1"
        self.stream_hash = "abc123"
        self.udiffVER_Ok = False
        self.badimg = 0
        self.pollInterval = 500
        self.connected = False
        self.image_name = self.get_property("image_name")
        self.image = HWR.get_hardware_repository().find_in_repository(self.image_name)
        self.update_state(BaseHardwareObjects.HardwareObjectState.READY)
        self._video_stream_process = None
        self._current_stream_size = (0, 0)

    def init(self):
        logging.getLogger("HWR").info("initializing camera object")
        if self.get_property("interval"):
            self.pollInterval = self.get_property("interval")
        self.stopper = False  # self.polling_timer(self.pollInterval, self.poll)
        gevent.spawn(self.poll)

    def udiffVersionChanged(self, value) -> None:
        if value == "MD2_2":
            print(("start polling MD camera with poll interval=", self.pollInterval))
        else:
            print(
                "stop polling the camera. This microdiff version does not support a camera"
            )
            self.stopper = True

    def connectToDevice(self) -> bool:
        self.connected = True
        return self.connected

    def poll(self) -> None:
        logging.getLogger("HWR").info("going to poll images")
        while not self.stopper:
            time.sleep(1)
            try:
                img = open(self.image, "rb").read()
                self.emit("imageReceived", img, 659, 493)
            except Exception:
                logging.getLogger("HWR").exception("Could not read image")

    def imageUpdated(self, value) -> None:
        print("<HW> got new image")
        print(value)

    def gammaExists(self) -> bool:
        return False

    def contrastExists(self) -> bool:
        return False

    def brightnessExists(self) -> bool:
        return False

    def gainExists(self) -> bool:
        return False

    def get_width(self) -> int:
        # return 768 #JN ,20140807,adapt the MD2 screen to mxCuBE2
        return 659

    def get_height(self) -> int:
        # return 576 # JN ,20140807,adapt the MD2 screen to mxCuBE2
        return 493

    def set_live(self, state) -> bool:
        self.liveState = state
        return True

    def get_last_image(self) -> Tuple[bytes, int, int]:
        image = Image.open(self.image)
        return image.tobytes(), image.size[0], image.size[1]

    def get_available_stream_sizes(self) -> List[Tuple[int, int]]:
        try:
            w, h = self.get_width(), self.get_height()
            video_sizes = [(w, h), (int(w / 2), int(h / 2)), (int(w / 4), int(h / 4))]
        except (ValueError, AttributeError):
            video_sizes = []

        return video_sizes

    def set_stream_size(self, w, h) -> None:
        self._current_stream_size = (int(w), int(h))

    def get_stream_size(self) -> Tuple[int, int, float]:
        width, height = self._current_stream_size
        scale = float(width) / self.get_width()
        return (width, height, scale)

    def clean_up(self):
        logging.getLogger("HWR").info("Shutting down video_stream...")
        os.kill(self._video_stream_process.pid, signal.SIGTERM)

    def start_video_stream_process(self) -> None:
        if (
            not self._video_stream_process
            or self._video_stream_process.poll() is not None
        ):
            self._video_stream_process = subprocess.Popen(
                [
                    "video-streamer",
                    "-uri",
                    "test",
                    "-hs",
                    "localhost",
                    "-p",
                    str(self._port),
                    "-of",
                    self._format,
                    "-q",
                    "4",
                    "-s",
                    ", ".join(map(str, self._current_stream_size)),
                    "-id",
                    self.stream_hash,
                ],
                close_fds=True,
                stdout=subprocess.DEVNULL,
            )

            atexit.register(self.clean_up)

    def stop_streaming(self) -> None:
        if self._video_stream_process:
            try:
                ps = [self._video_stream_process] + psutil.Process(
                    self._video_stream_process.pid
                ).children()
                for p in ps:
                    p.kill()
            except psutil.NoSuchProcess:
                pass

            self._video_stream_process = None

    def start_streaming(self, _format="MPEG1", size=(0, 0), port="8000") -> None:
        self._format = _format
        self._port = port

        if not size[0]:
            _s = int(self.get_width()), int(self.get_height())
        else:
            _s = int(size[0]), int(size[1])

        self.set_stream_size(_s[0], _s[1])
        self.start_video_stream_process()

    def restart_streaming(self, size) -> None:
        self.stop_streaming()
        self.start_streaming(self._format, size=size)



import logging
import time

import epics

from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector


class LNLSPilatusDet(AbstractDetector):
    DET_THRESHOLD = "det_threshols_energy"
    # DET_STATUS = 'det_status_message'
    DET_WAVELENGTH = "det_wavelength"
    DET_DETDIST = "det_detdist"
    DET_BEAM_X = "det_beam_x"
    DET_BEAM_Y = "det_beam_y"
    USER_BEAM_X = "user_beam_x"
    USER_BEAM_Y = "user_beam_y"
    DET_TRANSMISSION = "det_transmission"
    DET_START_ANGLE = "det_start_angle"
    DET_ANGLE_INCR = "det_angle_incr"

    def __init__(self, name):
        """
        Descript. :
        """
        AbstractDetector.__init__(self, name)

    def init(self):
        """
        Descript. :
        """
        AbstractDetector.init(self)

        # self.distance = 500
        self._temperature = 25
        self._humidity = 60
        self.actual_frame_rate = 50
        self._roi_modes_list = ("0", "C2", "C16")
        self._roi_mode = 0
        self._exposure_time_limits = [0.04, 60000]
        self.status = "ready"
        self.pv_status = epics.PV(self.getProperty("channel_status"))
        self.threshold = -1  # Starts with invalid value. To be set.
        self.wavelength = -1
        self.det_distance = -1
        self.beam_x = -1
        self.beam_y = -1
        self.default_beam_x = float(self.getProperty("default_beam_x"))
        self.default_beam_y = float(self.getProperty("default_beam_y"))

        self._distance_motor_hwobj = self.get_object_by_role("detector_distance")
        self.threshold = self.get_threshold_energy()

    def set_roi_mode(self, roi_mode):
        self._roi_mode = roi_mode
        self.emit("detectorModeChanged", (self._roi_mode,))

    def has_shutterless(self):
        """Returns always True"""
        return True

    def get_beam_position(self, distance=None, wavelength=None):
        """Get approx detector centre"""
        xval, yval = super(LNLSPilatusDet, self).get_beam_position(distance=distance)
        if None in (xval, yval):
            # default to Pilatus values
            xval = self.getProperty("width", 2463) / 2.0 + 0.4
            yval = self.getProperty("height", 2527) / 2.0 + 0.4
        return xval, yval

    def update_values(self):
        self.emit("detectorModeChanged", (self._roi_mode,))
        self.emit("temperatureChanged", (self._temperature, True))
        self.emit("humidityChanged", (self._humidity, True))
        self.emit("expTimeLimitsChanged", (self._exposure_time_limits,))
        self.emit("frameRateChanged", self.actual_frame_rate)
        self.emit("statusChanged", (self.status, "Ready"))

    def prepare_acquisition(self, *args, **kwargs):
        """
        Prepares detector for acquisition
        """
        return

    def last_image_saved(self):
        """
        Returns:
            str: path to last image
        """
        return

    def start_acquisition(self):
        """
        Starts acquisition
        """
        return

    def stop_acquisition(self):
        """
        Stops acquisition
        """
        return

    def get_threshold_energy(self):
        """
        Returns:
            float: threshold energy
        """
        value = float(self.get_channel_value(self.DET_THRESHOLD))
        return value

    def set_threshold_energy(self, energy):
        """
        Set threshold energy and returns whether it was successful or not.
        """
        try:
            float(energy)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting Pilatus threshold. Value must be float."
            )
            return False

        target_threshold = energy / 2
        if abs(self.get_threshold_energy() - target_threshold) < 0.0001:
            return True

        logging.getLogger("HWR").info("Setting Pilatus threshold...")
        for i in range(3):
            logging.getLogger("user_level_log").info("Setting Pilatus threshold...")

        self.set_channel_value(self.DET_THRESHOLD, target_threshold)

        # wait for threshold setting to be done
        time.sleep(2)
        # Using epics because we need 'as_string' option
        status = self.pv_status.get(as_string=True)
        logging.getLogger("HWR").info("Pilatus status: %s" % status)

        while status == "Setting threshold":
            logging.getLogger("HWR").info(
                "Pilatus status: %s (this may take a minute)..." % status
            )
            time.sleep(3)
            status = self.pv_status.get(as_string=True)

        self.threshold = self.get_threshold_energy()
        logging.getLogger("HWR").info(
            "Pilatus: current threshold is %s (target is %s)"
            % (self.threshold, target_threshold)
        )
        if status == "Camserver returned OK" and self.threshold == target_threshold:
            logging.getLogger("HWR").info("Pilatus status: %s" % status)
            logging.getLogger("HWR").info("Pilatus threshold successfully set.")
            return True

        logging.getLogger("HWR").error("Pilatus status: %s" % status)
        logging.getLogger("HWR").error(
            "Error while setting Pilatus threshold. Please, check the detector."
        )
        return False

    def get_wavelength(self):
        """
        Returns:
            float: wavelength
        """
        value = float(self.get_channel_value(self.DET_WAVELENGTH))
        return value

    def set_wavelength(self, wavelength):
        """
        Set wavelength and returns whether it was successful or not.
        """
        try:
            float(wavelength)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting Pilatus wavelength. Value must be float."
            )
            return False

        # if abs(self.wavelength - wavelength) < 0.0001:
        #    logging.getLogger("HWR").info(
        #        "Pilatus wavelength still okay."
        #    )
        #    return True

        # As the set of Pilatus wavelength, det dist and beam xy is fast,
        # there is no need to compare the target value with the current one.
        logging.getLogger("HWR").info("Setting Pilatus wavelength...")
        self.set_channel_value(self.DET_WAVELENGTH, wavelength)
        time.sleep(0.6)

        print("WAVELENGHT WAS SET: " + str(wavelength))
        self.wavelength = self.get_wavelength()
        print("WAVELENGHT GOT: " + str(self.wavelength))

        if abs(self.wavelength - wavelength) < 0.0001:
            logging.getLogger("HWR").info("Pilatus wavelength successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus wavelength. Please, check the detector."
        )
        return False

    def get_detector_distance(self):
        """
        Returns:
            float: detector distance
        """
        value = float(self.get_channel_value(self.DET_DETDIST))
        return value

    def set_detector_distance(self, det_distance):
        """
        Set detector distance and returns whether it was successful or not.
        """
        try:
            float(det_distance)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting Pilatus det distance. Value must be float."
            )
            return False

        # if abs(self.det_distance - det_distance) < 0.001:
        #    logging.getLogger("HWR").info(
        #        "Pilatus det distance still okay."
        #    )
        #    return True

        logging.getLogger("HWR").info("Setting Pilatus det distance...")
        self.set_channel_value(self.DET_DETDIST, det_distance)
        time.sleep(0.3)

        self.det_distance = self.get_detector_distance()

        if abs(self.det_distance - det_distance) < 0.001:
            logging.getLogger("HWR").info("Pilatus det distance successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus det distance. Please, check the detector."
        )
        return False

    def get_user_beam_x(self):
        """
        Returns:
            float: user beam x
        """
        value = float(self.get_channel_value(self.USER_BEAM_X))
        return value

    def get_beam_x(self):
        """
        Returns:
            float: detector beam x
        """
        value = float(self.get_channel_value(self.DET_BEAM_X))
        return value

    def set_beam_x(self, from_user=False, beam_x=None):
        """
        Set detector beam_x and returns whether it was successful or not.

        Beam X value can come from different sources. The priority (from
        high to low) is:
        * from_user
        * beam_x argument
        * default_beam_x (value set on xml file)
        """
        if from_user:
            logging.getLogger("HWR").info("Getting beam X from user...")
            beam_x = self.get_user_beam_x()

        if beam_x is None:
            if from_user:
                logging.getLogger("HWR").error(
                    "Could not get user beam X. Setting default value (from xml)."
                )
            beam_x = self.default_beam_x

        # if abs(self.beam_x - beam_x) == 0:
        #    logging.getLogger("HWR").info(
        #        "Pilatus beam X still okay."
        #    )
        #    return True

        logging.getLogger("HWR").info("Setting Pilatus beam X to {}...".format(beam_x))
        self.set_channel_value(self.DET_BEAM_X, beam_x)
        time.sleep(1)

        self.beam_x = self.get_beam_x()

        if float(self.beam_x) == float(beam_x):
            logging.getLogger("HWR").info("Pilatus det beam X successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus beam X. Please, check the detector."
        )
        return False

    def get_user_beam_y(self):
        """
        Returns:
            float: user beam y
        """
        value = float(self.get_channel_value(self.USER_BEAM_Y))
        return value

    def get_beam_y(self):
        """
        Returns:
            float: detector beam y
        """
        value = float(self.get_channel_value(self.DET_BEAM_Y))
        return value

    def set_beam_y(self, from_user=False, beam_y=None):
        """
        Set detector beam_y and returns whether it was successful or not.

        Beam Y value can come from different sources. The priority (from
        high to low) is:
        * from_user
        * beam_y argument
        * default_beam_y (value set on xml file)
        """
        if from_user:
            logging.getLogger("HWR").info("Getting beam Y from user...")
            beam_y = self.get_user_beam_y()

        if beam_y is None:
            if from_user:
                logging.getLogger("HWR").error(
                    "Could not get user beam Y. Setting default value (from xml)."
                )
            beam_y = self.default_beam_y

        # if abs(self.beam_y - beam_y) == 0:
        #    logging.getLogger("HWR").info(
        #        "Pilatus beam X still okay."
        #    )
        #    return True

        logging.getLogger("HWR").info("Setting Pilatus beam Y to {}...".format(beam_y))
        self.set_channel_value(self.DET_BEAM_Y, beam_y)
        time.sleep(1)

        self.beam_y = self.get_beam_y()

        if float(self.beam_y) == float(beam_y):
            logging.getLogger("HWR").info("Pilatus det beam Y successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus beam Y. Please, check the detector."
        )
        return False

    def get_transmission(self):
        """
        Returns:
            float: detector filter transmission value
        """
        value = float(self.get_channel_value(self.DET_TRANSMISSION))
        return value

    def set_transmission(self, transmission):
        """
        Set filter transmission and returns whether it was successful or not.
        """
        try:
            float(transmission)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting Pilatus transmission. Value must be float."
            )
            return False

        logging.getLogger("HWR").info(
            "Setting Pilatus transmission to {}...".format(transmission)
        )
        self.set_channel_value(self.DET_TRANSMISSION, transmission)
        time.sleep(0.3)

        self.transmission = self.get_transmission()

        if abs(self.transmission - transmission) < 0.0001:
            logging.getLogger("HWR").info("Pilatus transmission successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus transmission. Please, check the detector."
        )
        return False

    def get_start_angle(self):
        """
        Returns:
            float: detector start angle value
        """
        value = float(self.get_channel_value(self.DET_START_ANGLE))
        return value

    def set_start_angle(self, start_angle):
        """
        Set start angle and returns whether it was successful or not.
        """
        try:
            float(start_angle)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting Pilatus start angle. Value must be float."
            )
            return False

        logging.getLogger("HWR").info(
            "Setting Pilatus start angle to {}...".format(start_angle)
        )
        self.set_channel_value(self.DET_START_ANGLE, start_angle)
        time.sleep(3)

        self.start_angle = self.get_start_angle()

        if abs(self.start_angle - start_angle) < 0.0001:
            logging.getLogger("HWR").info("Pilatus start angle successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus start angle. Please, check the detector."
        )
        return False

    def get_angle_incr(self):
        """
        Returns:
            float: detector angle increment value
        """
        value = float(self.get_channel_value(self.DET_ANGLE_INCR))
        return value

    def set_angle_incr(self, angle_incr):
        """
        Set angle increment and returns whether it was successful or not.
        """
        try:
            float(angle_incr)
        except Exception as e:
            logging.getLogger("HWR").error(
                "Error while setting Pilatus angle increment. Value must be float."
            )
            return False

        logging.getLogger("HWR").info(
            "Setting Pilatus angle increment to {}...".format(angle_incr)
        )
        self.set_channel_value(self.DET_ANGLE_INCR, angle_incr)
        time.sleep(3)

        self.angle_incr = self.get_angle_incr()

        if abs(self.angle_incr - angle_incr) < 0.0001:
            logging.getLogger("HWR").info("Pilatus angle increment successfully set.")
            return True

        logging.getLogger("HWR").error(
            "Error while setting Pilatus angle increment. Please, check the detector."
        )
        return False



import logging
import sys
import time

import gevent
import PyTango.gevent

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR


class ID29HutchTrigger(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        BaseHardwareObjects.HardwareObject.__init__(self, name)

    def _do_polling(self):
        while True:
            try:
                self.poll()
            except Exception:
                sys.excepthook(*sys.exc_info())
            time.sleep(self.get_property("interval") / 1000.0 or 1)

    def init(self):
        try:
            self.device = PyTango.gevent.DeviceProxy(self.get_property("tangoname"))
        except PyTango.DevFailed as traceback:
            last_error = traceback[-1]
            logging.getLogger("HWR").error(
                "%s: %s", str(self.name()), last_error["desc"]
            )
            self.device = None

        self.pollingTask = None
        self.initialized = False
        self.__oldValue = None
        self.card = None
        self.channel = None

        PSSinfo = self.get_property("pss")
        try:
            self.card, self.channel = map(int, PSSinfo.split("/"))
        except Exception:
            logging.getLogger().error("%s: cannot find PSS number", self.name())
            return

        if self.device is not None:
            self.pollingTask = gevent.spawn(self._do_polling)
        self.connected()

    def hutchIsOpened(self):
        return self.hutch_opened

    def is_connected(self):
        return True

    def connected(self):
        self.emit("connected")

    def disconnected(self):
        self.emit("disconnected")

    def abort(self):
        pass

    def macro(self, entering_hutch, old={"dtox": None}):
        logging.info(
            "%s: %s hutch", self.id, "entering" if entering_hutch else "leaving"
        )
        dtox = HWR.beamline.detector.distance
        if not entering_hutch:
            if old["dtox"] is not None:
                dtox.set_value(old["dtox"])
            self.get_command_object("macro")(0)
        else:
            old["dtox"] = dtox.get_value()
            self.get_command_object("macro")(1)
            dtox.set_value(700)
        dtox.waitEndOfMove()

    def poll(self):
        a = self.device.GetInterlockState([self.card - 1, 2 * (self.channel - 1)])[0]
        b = self.device.GetInterlockState([self.card - 1, 2 * (self.channel - 1) + 1])[
            0
        ]
        value = a & b

        if value == self.__oldValue:
            return
        else:
            self.__oldValue = value

        self.value_changed(value)

    def value_changed(self, value, *args):
        if value == 0:
            if self.initialized:
                self.emit("hutchTrigger", (1,))
        elif value == 1 and self.initialized:
            self.emit("hutchTrigger", (0,))

        self.hutch_opened = 1 - value
        self.initialized = True



# -*- coding: utf-8 -*-
# from SimpleDevice2c import SimpleDevice
import logging
import math

from PyTango.gevent import DeviceProxy

from mxcubecore.BaseHardwareObjects import HardwareObject

# from Command.Tango import TangoChannel


class PX2Attenuator(HardwareObject):
    stateAttenuator = {
        "ALARM": "error",
        "OFF": "error",
        "RUNNING": "moving",
        "MOVING": "moving",
        "STANDBY": "ready",
        "UNKNOWN": "changed",
        "EXTRACT": "outlimits",
    }

    def __init__(self, name):
        super().__init__(name)

        self.labels = []
        self.attno = 0
        self.deviceOk = True
        self.nominal_value = None

    def init(self):
        #         cmdToggle = self.get_command_object('toggle')
        #         cmdToggle.connect_signal('connected', self.connected)
        #         cmdToggle.connect_signal('disconnected', self.disconnected)

        # Connect to device FP_Parser defined "tangoname" in the xml file
        try:
            # self.Attenuatordevice = SimpleDevice(self.get_property("tangoname"), verbose=False)
            self.Attenuatordevice = DeviceProxy(self.get_property("tangoname"))
        except Exception:
            self.errorDeviceInstance(self.get_property("tangoname"))

        try:
            # self.Attenuatordevice = SimpleDevice(self.get_property("tangoname"), verbose=False)
            self.Constdevice = DeviceProxy(self.get_property("tangoname_const"))
        except Exception:
            self.errorDeviceInstance(self.get_property("tangoname_const"))

        # Connect to device Primary slit horizontal defined "tangonamePs_h" in the
        # xml file
        try:
            # self.Ps_hdevice = SimpleDevice(self.get_property("tangonamePs_h"), verbose=False)
            self.Ps_hdevice = DeviceProxy(self.get_property("tangonamePs_h"))
        except Exception:
            self.errorDeviceInstance(self.get_property("tangonamePs_h"))

        # Connect to device Primary slit vertical defined "tangonamePs_v" in the
        # xml file
        try:
            # self.Ps_vdevice = SimpleDevice(self.get_property("tangonamePs_v"), verbose=False)
            self.Ps_vdevice = DeviceProxy(self.get_property("tangonamePs_v"))
        except Exception:
            self.errorDeviceInstance(self.get_property("tangonamePs_v"))

        if self.deviceOk:
            self.connected()

            self.chanAttState = self.get_channel_object("State")
            self.chanAttState.connect_signal("update", self.attStateChanged)
            self.chanAttFactor = self.get_channel_object("TrueTrans_FP")
            self.chanAttFactor.connect_signal("update", self.attFactorChanged)

    def getAtteConfig(self):
        return

    def getAttState(self):
        try:
            value1 = Ps_attenuatorPX2.stateAttenuator[self.Ps_hdevice.State().name]
            print(
                "State hslit : ",
                Ps_attenuatorPX2.stateAttenuator[self.Ps_hdevice.State().name],
            )
            value2 = Ps_attenuatorPX2.stateAttenuator[self.Ps_vdevice.State().name]
            print(
                "State vslit : ",
                Ps_attenuatorPX2.stateAttenuator[self.Ps_vdevice.State().name],
            )
            if value1 == "ready" and value2 == "ready":
                value = "ready"
            elif value1 == "error" or value2 == "error":
                value = "error"
            elif value1 == "moving" or value2 == "moving":
                value = "moving"
            elif value1 == "changed" or value == "changed":
                value = "changed"
            else:
                value = None
            logging.getLogger().debug("Attenuator state read from the device %s", value)

        except Exception:
            logging.getLogger("HWR").error(
                "%s getAttState : received value on channel is not a integer value",
                str(self.name()),
            )
            value = None
        return value

    def attStateChanged(self, channelValue):
        value = self.getAttState()
        logging.getLogger("HWR").error(
            "%s getAttState : new value is %s" % (str(self.name()), value)
        )
        self.emit("attStateChanged", (value,))

    def get_value(self):
        try:
            if (
                self.Attenuatordevice.TrueTrans_FP <= 100.0
            ):  # self.Attenuatordevice.Trans_FP  <= 100.0 :
                if self.nominal_value is not None:
                    value = self.nominal_value
                else:
                    value = float(self.Attenuatordevice.TrueTrans_FP) * 1.3587
            else:
                if self.nominal_value is not None:
                    value = self.nominal_value
                else:
                    value = float(self.Attenuatordevice.I_Trans) * 1.4646
            # Mettre une limite superieure car a une certaine ouverture de fentes on ne gagne plus rien en transmission
            # Trouver la valeur de transmission par mesure sur QBPM1 doit etre autour
            # de 120%
        except Exception:
            logging.getLogger("HWR").error(
                "%s get_value : received value on channel is not a float value",
                str(self.name()),
            )
            value = None
        return value

    def connected(self):
        self.set_is_ready(True)

    def disconnected(self):
        self.set_is_ready(False)

    def attFactorChanged(self, channelValue):
        try:
            logging.getLogger("HWR").info(
                "%s attFactorChanged : received value %s"
                % (str(self.name()), channelValue)
            )
            value = self.get_value()
        except Exception:
            logging.getLogger("HWR").error(
                "%s attFactorChanged : received value on channel is not a float value",
                str(self.name()),
            )
        else:
            logging.getLogger("HWR").info(
                "%s attFactorChanged : calculated value is %s"
                % (str(self.name()), value)
            )
            self.emit("attFactorChanged", (value,))

    def attToggleChanged(self, channelValue):
        try:
            value = int(channelValue)
        except Exception:
            logging.getLogger("HWR").error(
                "%s attToggleChanged : received value on channel is not a float value",
                str(self.name()),
            )
        else:
            self.emit("toggleFilter", (value,))

    def _set_value(self, value):
        self.nominal_value = float(value)
        try:
            if (
                self.Constdevice.FP_Area_FWHM <= 0.1
            ):  # Cas ou il n'y a pas de valeur dans le publisher PASSERELLE/CO/Primary_Slits
                logging.getLogger("HWR").error(
                    "Primary slits not correctly aligned", str(self.name())
                )
                self.Constdevice.FP_Area_FWHM = 0.5
                self.Constdevice.Ratio_FP_Gap = 0.5

            truevalue = (2.0 - math.sqrt(4 - 0.04 * value)) / 0.02
            print(" truevalue : ", truevalue)
            newGapFP_H = math.sqrt(
                (truevalue / 100.0)
                * self.Constdevice.FP_Area_FWHM
                / self.Constdevice.Ratio_FP_Gap
            )
            print(" Gap FP_H : ", newGapFP_H)
            self.Ps_hdevice.gap = newGapFP_H
            newGapFP_V = newGapFP_H * self.Constdevice.Ratio_FP_Gap
            print(" Gap FP_V : ", newGapFP_V)
            self.Ps_vdevice.gap = newGapFP_V
            # self.attFactorChanged(channelValue)
        except Exception:
            logging.getLogger("HWR").error(
                "%s set Transmission : received value on channel is not valid",
                str(self.name()),
            )
            value = None
        return value

    def toggle(self, value):
        return value

    def errorDeviceInstance(self, device):
        db = DeviceProxy("sys/database/dbds1")
        logging.getLogger().error(
            "Check Instance of Device server %s"
            % db.Dbget_deviceInfo(HardwareObject)[1][3]
        )
        self.sDisconnected()


def test_hwo(hwo):
    print("Atten. factor", hwo.get_value())



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"
__author__ = "Jan Meyer"
__email__ = "jan.meyer@desy.de"


import json
import traceback

import gevent

try:
    from httplib import HTTPConnection
except ImportError:
    from http.client import HTTPConnection

try:
    import redis

    redis_flag = True
except ImportError:
    traceback.print_exc()
    redis_flag = False

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice
from mxcubecore.utils.conversion import string_types
from mxcubecore.utils.qt_import import (
    QImage,
    QPixmap,
    QPoint,
)


class MjpgStreamVideo(AbstractVideoDevice):
    """
    Hardware object to capture images using mjpg-streamer
    and it's input_avt.so plugin for AVT Prosilica cameras.
    """

    # command / control types supported by mjpg-streamer
    CTRL_TYPE_INTEGER = 1
    CTRL_TYPE_BOOLEAN = 2
    CTRL_TYPE_MENU = 3
    CTRL_TYPE_BUTTON = 4

    # command destinations
    DEST_INPUT = 0
    DEST_OUTPUT = 1
    DEST_PROGRAM = 2

    # command groups
    IN_CMD_GROUP_GENERIC = 0
    IN_CMD_GROUP_RESOLUTION = 2
    IN_CMD_GROUP_JPEG_QUALITY = 3
    IN_CMD_GROUP_AVT_MISC = 32
    IN_CMD_GROUP_AVT_INFO = 33
    IN_CMD_GROUP_AVT_EXPOSURE = 34
    IN_CMD_GROUP_AVT_GAIN = 35
    IN_CMD_GROUP_AVT_LENS_DRIVE = 36
    IN_CMD_GROUP_AVT_IRIS = 37
    IN_CMD_GROUP_AVT_WHITE_BALANCE = 38
    IN_CMD_GROUP_AVT_DSP = 39
    IN_CMD_GROUP_AVT_IMAGE_FORMAT = 40
    IN_CMD_GROUP_AVT_IO = 41
    IN_CMD_GROUP_AVT_ACQUISITION = 42
    IN_CMD_GROUP_AVT_CONFIG_FILE = 43
    IN_CMD_GROUP_AVT_NETWORK = 44
    IN_CMD_GROUP_AVT_STATS = 45
    IN_CMD_GROUP_AVT_EVENTS = 46

    # commands
    # note: not every camera supports every command
    # mjpg-streamer only supports integer values for commands
    #   float values are therefore submitted as an integer value * 1000
    #   booleans will only accept 0 or 1
    #   menus have an integer identifier for every item
    #   commands will ignore the given value
    #   strings and events are unsupported yet
    # for more information see the "AVT Camera and Driver Attributes" manual
    IN_CMD_UPDATE_CONTROLS = (1, IN_CMD_GROUP_GENERIC)
    IN_CMD_JPEG_QUALITY = (1, IN_CMD_GROUP_JPEG_QUALITY)
    IN_CMD_AVT_ACQ_END_TRIGGER_EVENT = (6, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQ_END_TRIGGER_MODE = (7, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQ_REC_TRIGGER_EVENT = (8, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQ_REC_TRIGGER_MODE = (9, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQ_START_TRIGGER_EVENT = (10, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQ_START_TRIGGER_MODE = (11, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQUISITION_ABORT = (5, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQUISITION_FRAME_COUNT = (2, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQUISITION_MODE = (1, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQUISITION_START = (3, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_ACQUISITION_STOP = (4, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_BANDWIDTH_CTRL_MODE = (1, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_BINNING_X = (1, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    IN_CMD_AVT_BINNING_Y = (2, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    # String 2 IN_CMD_AVT_CAMERA_NAME = (0, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_CHUNK_MODE_ACTIVE = (2, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_CONFIG_FILE_INDEX = (1, IN_CMD_GROUP_AVT_CONFIG_FILE)
    IN_CMD_AVT_CONFIG_FILE_LOAD = (3, IN_CMD_GROUP_AVT_CONFIG_FILE)
    IN_CMD_AVT_CONFIG_FILE_POWER_UP = (2, IN_CMD_GROUP_AVT_CONFIG_FILE)
    IN_CMD_AVT_CONFIG_FILE_SAVE = (4, IN_CMD_GROUP_AVT_CONFIG_FILE)
    IN_CMD_AVT_DSP_SUBREGION_BOTTOM = (4, IN_CMD_GROUP_AVT_DSP)
    IN_CMD_AVT_DSP_SUBREGION_LEFT = (1, IN_CMD_GROUP_AVT_DSP)
    IN_CMD_AVT_DSP_SUBREGION_RIGHT = (3, IN_CMD_GROUP_AVT_DSP)
    IN_CMD_AVT_DSP_SUBREGION_TOP = (2, IN_CMD_GROUP_AVT_DSP)
    IN_CMD_AVT_DEFECT_MASK_COLUMN_ENABLE = (1, IN_CMD_GROUP_AVT_MISC)
    # String 14 IN_CMD_AVT_DEVICE_ETH_ADDRESS = (0, IN_CMD_GROUP_AVT_NETWORK)
    # String 3 IN_CMD_AVT_DEVICE_FIRMWARE_VERSION = (0, IN_CMD_GROUP_AVT_INFO)
    # String 15 IN_CMD_AVT_DEVICE_IP_ADDRESS = (0, IN_CMD_GROUP_AVT_NETWORK)
    # String 4 IN_CMD_AVT_DEVICE_MODEL_NAME = (0, IN_CMD_GROUP_AVT_INFO)
    # String 5 IN_CMD_AVT_DEVICE_PART_NUMBER = (0, IN_CMD_GROUP_AVT_INFO)
    # String 6 IN_CMD_AVT_DEVICE_SCAN_TYPE = (0, IN_CMD_GROUP_AVT_INFO)
    # String 7 IN_CMD_AVT_DEVICE_SERIAL_NUMBER = (0, IN_CMD_GROUP_AVT_INFO)
    # String 8 IN_CMD_AVT_DEVICE_VENDOR_NAME = (0, IN_CMD_GROUP_AVT_INFO)
    # Event IN_CMD_AVT_EVENT_ACQUISITION_END = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_ACQUISITION_RECORD_TRIGGER = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_ACQUISITION_START = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_ERROR = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_EXPOSURE_END = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_FRAME_TRIGGER = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_NOTIFICATION = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_OVERFLOW = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SELECTOR = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN1_FALL = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN1_RISE = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN2_FALL = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN2_RISE = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN3_FALL = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN3_RISE = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN4_FALL = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENT_SYNC_IN4_RISE = (0, IN_CMD_GROUP_AVT_EVENTS)
    # Event IN_CMD_AVT_EVENTS_ENABLE1 = (0, IN_CMD_GROUP_AVT_EVENTS)
    IN_CMD_AVT_EXPOSURE_AUTO_ADJUST_TOL = (3, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_AUTO_ALG = (4, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_AUTO_MAX = (5, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_AUTO_MIN = (6, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_AUTO_OUTLIERS = (7, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_AUTO_RATE = (8, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_AUTO_TARGET = (9, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_MODE = (2, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_EXPOSURE_VALUE = (1, IN_CMD_GROUP_AVT_EXPOSURE)
    IN_CMD_AVT_FIRMWARE_VER_BUILD = (9, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_FIRMWARE_VER_MAJOR = (10, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_FIRMWARE_VER_MINOR = (11, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_FRAMERATE = (12, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_FRAME_START_TRIGGER_DELAY = (14, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_FRAME_START_TRIGGER_EVENT = (15, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_FRAME_START_TRIGGER_MODE = (13, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_FRAME_START_TRIGGER_OVERLAP = (16, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_FRAME_START_TRIGGER_SOFTWARE = (17, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_GAIN_AUTO_ADJUST_TOL = (3, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_AUTO_MAX = (4, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_AUTO_MIN = (5, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_AUTO_OUTLIERS = (6, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_AUTO_RATE = (7, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_AUTO_TARGET = (8, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_MODE = (2, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GAIN_VALUE = (1, IN_CMD_GROUP_AVT_GAIN)
    IN_CMD_AVT_GVSP_LOOKBACK_WINDOW = (18, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_GVSP_RESEND_PERCENT = (19, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_GVSP_RETRIES = (20, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_GVSP_SOCKET_BUFFERS_COUNT = (21, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_GVSP_TIMEOUT = (22, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_HEARTBEAT_INTERVAL = (23, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_HEARTBEAT_TIMEOUT = (24, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_HEIGHT = (6, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    # String 16 IN_CMD_AVT_HOST_ETH_ADDRESS = (0, IN_CMD_GROUP_AVT_NETWORK)
    # String 17 IN_CMD_AVT_HOST_IP_ADDRESS = (0, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_IRIS_AUTOTARGET = (1, IN_CMD_GROUP_AVT_IRIS)
    IN_CMD_AVT_IRIS_MODE = (2, IN_CMD_GROUP_AVT_IRIS)
    IN_CMD_AVT_IRIS_VIDEO_LEVEL = (3, IN_CMD_GROUP_AVT_IRIS)
    IN_CMD_AVT_IRIS_VIDEO_LEVEL_MAX = (4, IN_CMD_GROUP_AVT_IRIS)
    IN_CMD_AVT_IRIS_VIDEO_LEVEL_MIN = (5, IN_CMD_GROUP_AVT_IRIS)
    IN_CMD_AVT_LENS_DRIVE_COMMAND = (1, IN_CMD_GROUP_AVT_LENS_DRIVE)
    IN_CMD_AVT_LENS_DRIVE_DURATION = (2, IN_CMD_GROUP_AVT_LENS_DRIVE)
    IN_CMD_AVT_LENS_VOLTAGE = (3, IN_CMD_GROUP_AVT_LENS_DRIVE)
    IN_CMD_AVT_LENS_VOLTAGE_CONTROL = (4, IN_CMD_GROUP_AVT_LENS_DRIVE)
    IN_CMD_AVT_MULTICAST_ENABLE = (25, IN_CMD_GROUP_AVT_NETWORK)
    # String 26 IN_CMD_AVT_MULTICAST_IP_ADDRESS = (0, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_NON_IMAGE_PAYLOAD_SIZE = (3, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_PACKET_SIZE = (27, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_PART_CLASS = (12, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_PART_NUMBER = (13, IN_CMD_GROUP_AVT_INFO)
    # String 14 IN_CMD_AVT_PART_REVISION = (0, IN_CMD_GROUP_AVT_INFO)
    # String 15 IN_CMD_AVT_PART_VERSION = (0, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_PAYLOAD_SIZE = (4, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_PIXEL_FORMAT = (7, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    IN_CMD_AVT_RECORDER_PRE_EVENT_COUNT = (18, IN_CMD_GROUP_AVT_ACQUISITION)
    IN_CMD_AVT_REGION_X = (3, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    IN_CMD_AVT_REGION_Y = (4, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    IN_CMD_AVT_SENSOR_BITS = (20, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_SENSOR_HEIGHT = (18, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_SENSOR_TYPE = (19, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_SENSOR_WIDTH = (17, IN_CMD_GROUP_AVT_INFO)
    # String 16 IN_CMD_AVT_SERIAL_NUMBER = (0, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_STAT_DRIVER_TYPE = (1, IN_CMD_GROUP_AVT_STATS)
    # String 2 IN_CMD_AVT_STAT_FILTER_VERSION = (0, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_FRAME_RATE = (3, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_FRAMES_COMPLETED = (4, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_FRAMES_DROPPED = (5, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_PACKETS_ERRONEOUS = (6, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_PACKETS_MISSED = (7, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_PACKETS_RECEIVED = (8, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_PACKETS_REQUESTED = (9, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STAT_PACKETS_RESENT = (10, IN_CMD_GROUP_AVT_STATS)
    IN_CMD_AVT_STREAM_BYTES_PER_SECOND = (6, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_STREAM_FRAME_RATE_CONSTRAIN = (5, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_STREAM_HOLD_CAPACITY = (8, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_STREAM_HOLD_ENABLE = (7, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_STROBE1_CONTROLLED_DURATION = (1, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_STROBE1_DELAY = (3, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_STROBE1_DURATION = (4, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_STROBE1_MODE = (2, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_IN1_GLITCH_FILTER = (5, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_IN2_GLITCH_FILTER = (6, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_IN_LEVELS = (7, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT1_INVERT = (9, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT1_MODE = (10, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT2_INVERT = (11, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT2_MODE = (12, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT3_INVERT = (13, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT3_MODE = (14, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT4_INVERT = (15, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT4_MODE = (16, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_SYNC_OUT_GPO_LEVELS = (8, IN_CMD_GROUP_AVT_IO)
    IN_CMD_AVT_TIME_STAMP_FREQUENCY = (9, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_TIME_STAMP_RESET = (13, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_TIME_STAMP_VALUE_HI = (11, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_TIME_STAMP_VALUE_LATCH = (12, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_TIME_STAMP_VALUE_LO = (10, IN_CMD_GROUP_AVT_NETWORK)
    IN_CMD_AVT_TOTAL_BYTES_PER_FRAME = (8, IN_CMD_GROUP_AVT_IMAGE_FORMAT)
    IN_CMD_AVT_UNIQUE_ID = (1, IN_CMD_GROUP_AVT_INFO)
    IN_CMD_AVT_VSUB_VALUE = (2, IN_CMD_GROUP_AVT_MISC)
    IN_CMD_AVT_WHITEBAL_AUTO_ADJUST_TOL = (4, IN_CMD_GROUP_AVT_WHITE_BALANCE)
    IN_CMD_AVT_WHITEBAL_AUTO_RATE = (5, IN_CMD_GROUP_AVT_WHITE_BALANCE)
    IN_CMD_AVT_WHITEBAL_MODE = (3, IN_CMD_GROUP_AVT_WHITE_BALANCE)
    IN_CMD_AVT_WHITEBAL_VALUE_BLUE = (2, IN_CMD_GROUP_AVT_WHITE_BALANCE)
    IN_CMD_AVT_WHITEBAL_VALUE_RED = (1, IN_CMD_GROUP_AVT_WHITE_BALANCE)
    IN_CMD_AVT_WIDTH = (5, IN_CMD_GROUP_AVT_IMAGE_FORMAT)

    def __init__(self, name):
        """
        Descript. :
        """
        AbstractVideoDevice.__init__(self, name)
        self.force_update = None
        self.sensor_dimensions = None
        self.image_dimensions = None
        self.image_polling = None
        self.image_type = None
        self.image = None
        self.sleep_time = 1
        self.host = None
        self.port = None
        self.path = "/"
        self.plugin = 0
        self.update_controls = None
        self.input_avt = None
        self.last_jpeg = None
        self.changing_pars = False
        if redis_flag:
            self.redis = redis.StrictRedis()
        else:
            self.redis = False

    def init(self):
        """
        Descript. :
        """
        self.sleep_time = self.get_property("interval")

        hw_width = self.get_property("width")
        hw_height = self.get_property("height")
        scale = self.get_property("scale", 1)

        self.display_width = hw_width * scale
        self.display_height = hw_height * scale

        self.log.debug(
            " MJPG video - width=%s,height=%s - scale=%s" % (hw_width, hw_height, scale)
        )
        self.hw_dimensions = (hw_width, hw_height)
        self.image_dimensions = (self.display_width, self.display_height)
        self.scale = scale

        self.host = self.get_property("host")
        self.port = int(self.get_property("port"))

        self.standard_fliph = bool(self.get_property("fliph"))
        self.standard_flipv = bool(self.get_property("flipv"))
        self.standard_offsetx = 0
        self.standard_offsety = 0

        # parameters for overview camera
        self.overview_host = self.get_property("overview_host")
        self.overview_port = int(self.get_property("overview_port"))
        self.overview_fliph = bool(self.get_property("overview_fliph"))
        self.overview_flipv = bool(self.get_property("overview_flipv"))
        self.overview_offsetx = bool(self.get_property("overview_offsetx"))
        self.overview_offsety = bool(self.get_property("overview_offsety"))

        self.path = "/"
        self.plugin = 0

        self.using_overview = False

        self.update_controls = self.has_update_controls()
        self.input_avt = self.is_input_avt()

        self.image = self.get_new_image()

        if self.input_avt:
            sensor_info = self.get_cmd_info(self.IN_CMD_AVT_SENSOR_WIDTH)
            if sensor_info:
                sensor_width = int(sensor_info["value"])
            sensor_info = self.get_cmd_info(self.IN_CMD_AVT_SENSOR_HEIGHT)
            if sensor_info:
                sensor_height = int(sensor_info["value"])
            self.sensor_dimensions = (sensor_width, sensor_height)

        # self.is_ready()
        self.set_zoom(0)  # overview camera

    def http_get(self, query, host=None, port=None, path=None):
        """Sends HTTP GET requests and returns the answer.

        Keyword arguments:
        query -- string appended to the end of the requested URL
        host -- queried IP or hostname (default host of the MjpgStream instance)
        port -- queried port number (default port of the MjpgStream instance)
        path -- queried path (default path of the MjpgStream instance)

        Return value:
        the HTTP answer content or None on error

        """
        if self.using_overview is True:
            host = self.overview_host
            port = self.overview_port
            path = self.path
        else:
            host = self.host
            port = self.port
            path = self.path

        # send get request and return response
        http = HTTPConnection(host, port, timeout=3)
        # self.log.debug("MjpgStreamVideo: %s:%s - sending %s / %s " % (host,port,path,query))

        try:
            http.request("GET", path + query)
            response = http.getresponse()
        except Exception:
            self.log.error(
                "MjpgStreamVideo: Connection to http://{0}:{1}{2}{3} refused".format(
                    host, port, path, query
                )
            )
            return None
        if response.status != 200:
            self.log.error(
                "MjpgStreamVideo: Error {0}, {1}".format(
                    response.status, response.reason
                )
            )
            return None
        data = response.read()
        http.close()
        return data

    def send_cmd(self, value, cmd, group=None, plugin=None, dest=None):
        """Sends a command to mjpg-streamer.

        Keyword arguments:
        value -- command parameter as integer or item name as string if the command is of enumeration type
        cmd -- command id number or tuple constant
        group -- command group number, leave it at None if a tuple is given as cmd (default None)
        plugin -- plugin number (default plugin of the MjpgStream instance)
        dest -- command destination  (default MjpgStream.DEST_INPUT)

        """
        if isinstance(cmd, tuple) and group is None:
            group = cmd[1]
            cmd = cmd[0]
        elif isinstance(cmd, tuple):
            cmd = cmd[0]
        if group is None:
            return None
        cmd = str(int(cmd))
        group = str(int(group))
        try:
            value = str(int(value))
        except Exception:
            option = value
            value = None
            if isinstance(option, string_types):
                info = self.get_cmd_info(cmd, group)
                if info and "menu" in info and option in info["menu"].values():
                    value = str(
                        [k for k, v in info["menu"].items() if (v == option)][0]
                    )
            if value is None:
                return None
        if plugin is None:
            plugin = str(self.plugin)
        else:
            plugin = str(int(plugin))
        if dest is None:
            dest = str(self.DEST_INPUT)
        else:
            dest = str(int(dest))
        # send request
        self.http_get(
            "?action=command&id="
            + cmd
            + "&dest="
            + dest
            + "&group="
            + group
            + "&value="
            + value
            + "&plugin="
            + plugin
        )

    def has_cmd(self, cmd, group=None, plugin=None, dest=None):
        """Checks whether a command with the given id and group is known by the specified plugin.

        Keyword arguments:
        cmd -- command id number or tuple constant
        group -- command group number, leave it at None if a tuple is given as cmd (default None)
        plugin -- plugin number (default plugin of the MjpgStream instance)
        dest -- command destination  (default MjpgStream.DEST_INPUT)

        Return value:
        True if so, False if not or the connection was refused

        """
        data = self.get_cmd_info(cmd, group, plugin, dest)
        if not data is None:
            return True
        return False

    def get_cmd_info(self, cmd, group=None, plugin=None, dest=None):
        """Returns a dictionary with information on the queried command.

        Keyword arguments:
        cmd -- command id number or tuple constant
        group -- command group number, leave it at None if a tuple is given as cmd (default None)
        plugin -- plugin number (default plugin of the MjpgStream instance)
        dest -- command destination  (default MjpgStream.DEST_INPUT)

        Return value:
        dictionary containing the following items ("menu" only for menu commands):
        "name", "id", "type", "min", "max", "step", "default", "value", "dest", "flags", "group", "menu"
        or None on error

        """
        if isinstance(cmd, tuple) and group is None:
            group = cmd[1]
            cmd = cmd[0]
        elif isinstance(cmd, tuple):
            cmd = cmd[0]
        if group is None:
            return None
        if plugin is None:
            plugin = self.plugin
        else:
            plugin = str(int(plugin))
        if dest is None or (dest != self.DEST_INPUT and dest != self.DEST_OUTPUT):
            dest = self.DEST_INPUT
        if self.update_controls:
            self.send_cmd(group, self.IN_CMD_UPDATE_CONTROLS, plugin, dest)
        # get list of controls and search for the matching one
        data = self.get_controls(plugin, dest)
        if data is not None:
            for info in data:
                if int(info["group"]) == int(group) and int(info["id"]) == int(cmd):
                    return info
        return None

    def get_controls(self, plugin=None, dest=None):
        """Returns a list with information on all commands supported by the
        plugin. If DEST_PROGRAM is given for dest, a list with information on
        all loaded plugins is returned.

        Keyword arguments:
        plugin -- plugin number (default plugin of the MjpgStream instance)
        dest -- command destination  (default MjpgStream.DEST_INPUT)

        Return value:
        depends on destination plugin. For input_avt.so a list with all commands
        supported by the connected camera is returned - q.v. get_cmd_info().

        """
        if plugin is None:
            plugin = str(self.plugin)
        else:
            plugin = str(int(plugin))
        if dest is None:
            dest = self.DEST_INPUT
        else:
            dest = int(dest)
        query = None
        if dest == self.DEST_INPUT:
            query = "input_" + plugin + ".json"
        elif dest == self.DEST_OUTPUT:
            query = "output_" + plugin + ".json"
        elif dest == self.DEST_PROGRAM:
            query = "program.json"
        # fetch json from server, decode it into a python object and return it
        if query is not None:
            data = self.http_get(query)
            if data is not None:
                data = json.loads(data.decode("utf-8"))
                if dest != self.DEST_PROGRAM and "controls" in data:
                    data = data["controls"]
                return data
        return None

    def has_update_controls(self):
        """Checks if the default plugin for this instance knows the UpdateControls command.

        Return value:
        True if so, False if not and None if the connection was refused

        """
        data = self.get_cmd_info(self.IN_CMD_UPDATE_CONTROLS)
        if data is None:
            return None
        if data["name"] == "UpdateControls":
            return True
        return False

    def is_input_avt(self):
        """Checks if the default plugin for this instance is input_avt.so.

        Return value:
        True if so, False if not and None if the connection was refused

        """
        data = self.get_controls(0, self.DEST_PROGRAM)
        if data is None:
            return None
        if "inputs" in data:
            for info in data["inputs"]:
                if info["name"][-12:] == "input_avt.so":
                    return True
        return False

    def start_camera(self):
        if self.image_polling is None:
            self.image_polling = gevent.spawn(
                self._do_imagePolling, 1.0 / self.sleep_time
            )

    def get_image_dimensions(self):
        return self.image_dimensions

    def get_scale(self):
        return self.scale

    def contrast_exists(self):
        """
        Descript. :
        """
        return

    def set_contrast(self, contrast):
        """
        Descript. :
        """
        return

    def get_contrast(self):
        """
        Descript. :
        """
        return

    def set_contrast_auto(self, state=True):
        """
        Descript. :
        """
        return

    def get_contrast_auto(self):
        """
        Descript. :
        """
        return

    def get_contrast_min_max(self):
        """
        Descript. :
        """
        return

    def brightnessExists(self):
        """
        Descript. :
        """
        return

    def set_brightness(self, brightness):
        """
        Descript. :
        """
        return

    def get_brightness(self):
        """
        Descript. :
        """
        return

    def set_brightness_auto(self, state=True):
        """
        Descript. :
        """
        return

    def get_brightness_auto(self):
        """
        Descript. :
        """
        return

    def get_brightness_min_max(self):
        """
        Descript. :
        """
        return

    def gain_exists(self):
        """
        Descript. :
        """
        return self.has_cmd(self.IN_CMD_AVT_GAIN_VALUE)

    def set_gain(self, gain):
        """
        Descript. :
        """
        # self.send_cmd("Manual", self.IN_CMD_AVT_GAIN_MODE) # gain mode manual
        self.send_cmd(gain, self.IN_CMD_AVT_GAIN_VALUE)

    def get_gain(self):
        """
        Descript. :
        """
        info = self.get_cmd_info(self.IN_CMD_AVT_GAIN_VALUE)
        if info is not None:
            return float(info["value"])
        return

    def set_gain_auto(self, state=True):
        """
        Descript. :
        """
        if bool(state):
            self.send_cmd("Auto", self.IN_CMD_AVT_GAIN_MODE)
        else:
            self.send_cmd("Manual", self.IN_CMD_AVT_GAIN_MODE)

    def get_gain_auto(self):
        """
        Descript. :
        """
        value = None
        info = self.get_cmd_info(self.IN_CMD_AVT_GAIN_MODE)
        if info is not None:
            value = info["menu"][info["value"]] == "Auto"
        return value

    def get_gain_min_max(self):
        """
        Descript. :
        """
        info = self.get_cmd_info(self.IN_CMD_AVT_GAIN_VALUE)
        if info is not None:
            return (float(info["min"]), float(info["max"]))
        return

    def gamma_exists(self):
        """
        Descript. :
        """
        return

    def set_gamma(self, gamma):
        """
        Descript. :
        """
        return

    def get_gamma(self):
        """
        Descript. :
        """
        return

    def set_gamma_auto(self, state=True):
        """
        Descript. :
        """
        return

    def get_gamma_auto(self):
        """
        Descript. :
        """
        return

    def get_gamma_min_max(self):
        """
        Descript. :
        """
        return (0, 1)

    def exposure_time_exists(self):
        """
        Descript. :
        """
        return self.has_cmd(self.IN_CMD_AVT_EXPOSURE_VALUE)

    def set_exposure_time(self, gain):
        """
        Descript. :
        """
        # self.send_cmd("Manual", self.IN_CMD_AVT_EXPOSURE_MODE) # gain mode manual
        self.send_cmd(gain, self.IN_CMD_AVT_EXPOSURE_VALUE)

    def get_exposure_time(self):
        """
        Descript. :
        """
        info = self.get_cmd_info(self.IN_CMD_AVT_EXPOSURE_VALUE)
        if info is not None:
            return float(info["value"])
        return

    def set_exposure_time_auto(self, state=True):
        """
        Descript. :
        """
        if bool(state):
            self.send_cmd("Auto", self.IN_CMD_AVT_EXPOSURE_MODE)
        else:
            self.send_cmd("Manual", self.IN_CMD_AVT_EXPOSURE_MODE)

    def get_exposure_time_auto(self):
        """
        Descript. :
        """
        value = None
        info = self.get_cmd_info(self.IN_CMD_AVT_EXPOSURE_MODE)
        if info is not None:
            value = info["menu"][info["value"]] == "Auto"
        return value

    def get_exposure_time_min_max(self):
        """
        Descript. :
        """
        info = self.get_cmd_info(self.IN_CMD_AVT_EXPOSURE_VALUE)
        if info is not None:
            return (float(info["min"]), float(info["max"]))
        return

    def zoom_exists(self):
        """
        Descript. : True if the device supports digital zooming.
        """
        return self.input_avt and self.sensor_dimensions is not None

    def set_zoom(self, zoom):
        """
        Descript. : Sets digital zoom factor.
        """
        self.changing_pars = True

        if zoom == 0:
            self.using_overview = True
            width = self.sensor_dimensions[0] - 200
            height = self.sensor_dimensions[1] - 200

            offx, offy = self.overview_offsetx, self.overview_offsety

            pos_x = 60
            pos_y = 232
        else:
            self.using_overview = False

            limits = self.get_zoom_min_max()
            if zoom < limits[0] or zoom > limits[1]:
                return

            display_dimensions = self.image_dimensions
            width = display_dimensions[0] / zoom
            height = display_dimensions[1] / zoom

            pos_x = (self.sensor_dimensions[0] - width) / 2
            pos_y = (self.sensor_dimensions[1] - height) / 2

            pos_x = int(pos_x)
            pos_y = int(pos_y)

            width = int(width)
            height = int(height)

        self.send_cmd(1, self.IN_CMD_AVT_BINNING_X)
        gevent.sleep(0.1)
        self.send_cmd(1, self.IN_CMD_AVT_BINNING_Y)
        gevent.sleep(0.1)

        for i in range(3):  # try to program it three times
            if pos_x == 0:
                self.send_cmd(pos_x, self.IN_CMD_AVT_REGION_X)
                gevent.sleep(0.01)
            if pos_y == 0:
                self.send_cmd(pos_y, self.IN_CMD_AVT_REGION_Y)
                gevent.sleep(0.01)

            self.send_cmd(width, self.IN_CMD_AVT_WIDTH)
            gevent.sleep(0.01)
            self.send_cmd(height, self.IN_CMD_AVT_HEIGHT)
            gevent.sleep(0.01)

            if pos_y > 0:
                self.send_cmd(pos_y, self.IN_CMD_AVT_REGION_Y)
                gevent.sleep(0.01)
            if pos_x > 0:
                self.send_cmd(pos_x, self.IN_CMD_AVT_REGION_X)
                gevent.sleep(0.01)

            x_i = int(self.get_cmd_info(self.IN_CMD_AVT_REGION_X)["value"])
            y_i = int(self.get_cmd_info(self.IN_CMD_AVT_REGION_Y)["value"])
            w_i = int(self.get_cmd_info(self.IN_CMD_AVT_WIDTH)["value"])
            h_i = int(self.get_cmd_info(self.IN_CMD_AVT_HEIGHT)["value"])

            self.emit("zoomChanged", zoom)

            if (
                abs(x_i - pos_x) > 3
                or abs(y_i - pos_y) > 3
                or abs(w_i - width) > 3
                or abs(h_i - height) > 3
            ):
                self.log.debug(" - trying to program zoom again")
                gevent.sleep(0.1)
                continue
            else:
                break

        self.changing_pars = False
        zoom = self.get_zoom()
        if redis_flag:
            self.redis.set("zoom", zoom)
        self.emit("zoomChanged", zoom)

    def get_zoom(self):
        """
        Descript. : Returns the digital zoom factor.
        """
        if self.using_overview:
            return 0

        info = self.get_cmd_info(self.IN_CMD_AVT_WIDTH)
        if info is not None:
            return self.image_dimensions[0] / float(info["value"])

        return None

    def get_zoom_min_max(self):
        """
        Descript. : Returns the limits for the digital zoom factor.
                    The upper limit is arbitrarily defined.
        """
        return (self.image_dimensions[0] / float(self.sensor_dimensions[0]), 2)

    def set_live(self, mode):
        """
        Descript. :
        """
        return

    def get_width(self):
        """
        Descript. :
        """
        return self.image_dimensions[0]

    def get_height(self):
        """
        Descript. :
        """
        return self.image_dimensions[1]

    def get_new_image(self):
        """
        Descript. : reads new image, flips it if necessary and returns the
                    result or None on error
        """
        if self.using_overview:
            fliph, flipv = self.overview_fliph, self.overview_flipv
            offx, offy = self.overview_offsetx, self.overview_offsety
        else:
            fliph, flipv = self.standard_fliph, self.standard_flipv
            offx, offy = self.standard_offsetx, self.standard_offsety

        image = self.http_get("?action=snapshot")
        if image is not None:
            self.last_jpeg = image
            if redis_flag:
                self.redis.set("last_image_data", image)
            return QImage.fromData(image).mirrored(fliph, flipv)
        return None

    def refresh_video(self):
        """
        Descript. : reads new image into member variable, scales it and emits
                    imageReceived event. Added for compatibility.
        """
        image = self.get_new_image()
        if image is not None:
            image = image.scaled(self.display_width, self.display_height)
            # image.setOffset(QPoint(300,300))
            self.image = QPixmap.fromImage(
                image.scaled(self.display_width, self.display_height)
            )
            self.emit("imageReceived", self.image)

    def take_snapshot(self, filename, bw=False):
        """
        Descript. : calls get_new_image() and saves the result
        """
        try:
            qimage = self.get_new_image()
            # TODO convert to grayscale
            # if bw:
            #    qimage.setNumColors(0)
            qimage.save(filename, "PNG")
        except Exception:
            self.log.error("MjpgStreamVideo: unable to save snapshot: %s" % filename)

    def _do_imagePolling(self, sleep_time):
        """
        Descript. : worker method
        """
        while True:
            while self.changing_pars:
                self.log.debug("  / not reading image. busy changing pars")
                gevent.sleep(sleep_time)
                continue

            image = self.get_new_image()
            if image is not None:
                self.image = QPixmap.fromImage(
                    image.scaled(int(self.display_width), int(self.display_height))
                )
                self.emit("imageReceived", self.image)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
ALBAZoomMotorAutoBrightness

[Description]
Hardware Object is used to manipulate the zoom IOR and the
paired backlight intensity (slave IOR)

[Channels]
- None

[Commands]

[Emitted signals]
- stateChanged
- predefinedPositionChanged

[Functions]
- None

[Included Hardware Objects]
- zoom
- blight

Example Hardware Object XML file :
==================================
<object class="ALBAZoomMotorAutoBrightness">
  <object role="zoom" hwrid="/zoom"></object>
  <object role="blight" hwrid="/blight"></object>
</object>
"""

import logging

from mxcubecore import BaseHardwareObjects
from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor

__author__ = "Jordi Andreu"
__credits__ = ["MXCuBE collaboration"]

__version__ = "2.2."
__maintainer__ = "Jordi Andreu"
__email__ = "jandreu[at]cells.es"
__status__ = "Draft"


class ALBAZoomMotorAutoBrightness(BaseHardwareObjects.HardwareObject, AbstractMotor):
    INIT, FAULT, READY, MOVING, ONLIMIT = range(5)

    def __init__(self, name):
        super().__init__(name)

    def init(self):
        logging.getLogger("HWR").debug("Initializing zoom motor autobrightness IOR")

        self.zoom = self.get_object_by_role("zoom")
        self.blight = self.get_object_by_role("blight")

        self.zoom.positionChannel.connect_signal("update", self.positionChanged)
        self.zoom.stateChannel.connect_signal("update", self.stateChanged)

    def get_predefined_positions_list(self):
        retlist = self.zoom.get_predefined_positions_list()
        logging.getLogger("HWR").debug("Zoom positions list: %s" % repr(retlist))
        return retlist

    def moveToPosition(self, posno):
        # no = posno.split()[0]
        # logging.getLogger("HWR").debug("Moving to position %s" % no)

        # self.blight.moveToPosition(posno)
        self.zoom.moveToPosition(posno)
        state = self.zoom.get_state()
        # state = self.positionChannel.set_value(int(no))

    def motorIsMoving(self):
        return self.zoom.motorIsMoving()

    #        if str(self.get_state()) == "MOVING":
    #             return True
    #        else:
    #             return False

    def get_limits(self):
        # return (1,12)
        return self.zoom.get_limits()

    def get_state(self):
        #        state = self.stateChannel.get_value()
        #        curr_pos = self.get_value()
        #        if state == PyTango.DevState.ON:
        #             return ALBAZoomMotor.READY
        #        elif state == PyTango.DevState.MOVING or state == PyTango.DevState.RUNNING:
        #             return ALBAZoomMotor.MOVING
        #        elif curr_pos in self.get_limits():
        #             return ALBAZoomMotor.ONLIMIT
        #        else:
        #             return ALBAZoomMotor.FAULT
        #        return state
        return self.zoom.get_state()

    def get_value(self):
        return self.zoom.get_value()

    def get_current_position_name(self):
        #        n = int(self.positionChannel.get_value())
        #        value = "%s z%s" % (n, n)
        #        logging.getLogger("HWR").debug("get_current_position_name: %s" % repr(value))
        #        return value
        return self.zoom.get_current_position_name()

    def stateChanged(self, state):
        logging.getLogger("HWR").debug("stateChanged emitted: %s" % state)
        self.emit("stateChanged", (self.get_state(),))

    def positionChanged(self, currentposition):
        currentposition = self.get_current_position_name()
        logging.getLogger("HWR").debug(
            "predefinedPositionChanged emitted: %s" % currentposition
        )
        # Update light brightness step-by-step
        posno = currentposition.split()[0]
        logging.getLogger("HWR").debug("Moving brightness to: %s" % posno)

        self.blight.moveToPosition(posno)

        self.emit("predefinedPositionChanged", (currentposition, 0))

    def is_ready(self):
        state = self.get_state()
        return state == ALBAZoomMotorAutoBrightness.READY


def test():
    from mxcubecore import HardwareRepository as HWR

    hwr = HWR.get_hardware_repository()
    hwr.connect()

    zoom = hwr.get_hardware_object("/zoom-auto-brightness")

    print(type(zoom.get_state()))

    print("     Zoom position is : ", zoom.get_value())
    print("Zoom position name is : ", zoom.get_current_position_name())
    print("               Moving : ", zoom.motorIsMoving())
    print("                State : ", zoom.get_state())
    print("            Positions : ", zoom.get_predefined_positions_list())


if __name__ == "__main__":
    test()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractEnergy import AbstractEnergy

log = logging.getLogger("HWR")


class P11Energy(AbstractEnergy):
    _default_energy = 12.0

    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.chan_energy = self.get_channel_object("chanEnergy")
        if self.chan_energy is not None:
            self.chan_energy.connect_signal("update", self.energy_position_changed)

        self.chan_status = self.get_channel_object("chanStatus")
        if self.chan_status is not None:
            self.chan_status.connect_signal("update", self.energy_state_changed)

        self.chan_autobrake = self.get_channel_object("chanAutoBrake")
        limits = self.get_property("limits", None)

        try:
            limits = list(map(float, limits.split(",")))
        except Exception as e:
            log.error("P11Transmission - cannot parse limits: {}".format(str(e)))
            limits = None

        if limits is None:
            log.error(
                "P11Transmission - Cannot read LIMITS from configuration xml file.  Check values"
            )
            return
        else:
            self.set_limits(limits)

        self.energy_state_changed()
        self.energy_position_changed()

    def re_emit_values(self):
        self.emit("valueChanged", (self._nominal_value))
        self.emit("energyChanged", (self._nominal_value, self._wavelength_value))

    def is_ready(self):
        return self._state == self.STATES.READY

    def energy_state_changed(self, state=None):
        if state is None:
            state = self.chan_status.get_value()

        _state = str(state)

        if _state == "ON":
            self._state = self.STATES.READY
        elif _state == "MOVING":
            self._state = self.STATES.BUSY
        elif _state == "DISABLED":
            self._state = self.STATES.OFF
        else:
            self._state = self.STATES.FAULT

        self.emit("stateChanged", self._state)

    def energy_position_changed(self, pos=None):
        """
        Event called when energy value has been changed
        :param pos: float
        :return:
        """
        if pos is None:
            pos = self.chan_energy.get_value()

        energy = pos / 1000.0

        if self._nominal_value is None or abs(energy - self._nominal_value) > 1e-3:
            self._nominal_value = energy
            self._wavelength_value = 12.3984 / energy
            if self._wavelength_value is not None:
                self.emit(
                    "energyChanged", (self._nominal_value, self._wavelength_value)
                )
                self.emit("valueChanged", (self._nominal_value,))

    def _set_value(self, value):
        """
        Implementation pending
        """
        prog_value = value * 1000
        self.chan_autobrake.set_value(True)
        if self.get_state() == self.STATES.READY:
            current_value = self.get_value()
            if abs(current_value - value) < 0.01:
                self.log.debug(
                    "The difference between the current and desired energy values is less than 0.01. No change made."
                )
            else:
                self.log.debug("Programming ENERGY to %s eV" % prog_value)
                self.chan_energy.set_value(prog_value)
                self.log.debug(f"Energy value set to {value} keV.")

    def get_value(self):
        """
        Returns current energy in keV
        :return: float
        """

        value = self._default_energy
        if self.chan_energy is not None:
            try:
                value = self.chan_energy.get_value()
                return value / 1000
            except Exception:
                logging.getLogger("HWR").exception(
                    "Energy: could not read current energy"
                )
                return None
        return value


def test_hwo(hwo):
    print("Energy is: {0} keV".format(hwo.get_value()))



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model import queue_model_objects
from mxcubecore.queue_entry.base_queue_entry import (
    QUEUE_ENTRY_STATUS,
    BaseQueueEntry,
)
from mxcubecore.queue_entry.data_collection import DataCollectionQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class CharacterisationGroupQueueEntry(BaseQueueEntry):
    """
    Used to group (couple) a CollectionQueueEntry and a
    CharacterisationQueueEntry, creating a virtual entry for characterisation.
    """

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):
        BaseQueueEntry.__init__(self, view, data_model, view_set_queue_entry)
        self.dc_qe = None
        self.char_qe = None
        self.in_queue = False

    def execute(self):
        BaseQueueEntry.execute(self)

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)
        char = self.get_data_model()
        reference_image_collection = char.reference_image_collection

        # Trick to make sure that the reference collection has a sample.
        reference_image_collection._parent = char.get_parent()

        gid = self.get_data_model().get_parent().lims_group_id
        reference_image_collection.lims_group_id = gid

        # Enqueue the reference collection and the characterisation routine.
        dc_qe = DataCollectionQueueEntry(
            self.get_view(), reference_image_collection, view_set_queue_entry=False
        )
        dc_qe.set_enabled(True)
        dc_qe.in_queue = self.in_queue
        self.enqueue(dc_qe)
        self.dc_qe = dc_qe
        if char.run_characterisation:
            try:
                char_qe = CharacterisationQueueEntry(
                    self.get_view(), char, view_set_queue_entry=False
                )
            except Exception as ex:
                logging.getLogger("HWR").exception(
                    "Could not create CharacterisationQueueEntry"
                )
                self.char_qe = None
            else:
                char_qe.set_enabled(True)
                self.enqueue(char_qe)
                self.char_qe = char_qe

    def post_execute(self):
        if self.char_qe:
            self.status = self.char_qe.status
        else:
            self.status = self.dc_qe.status
        BaseQueueEntry.post_execute(self)


class CharacterisationQueueEntry(BaseQueueEntry):
    """
    Defines the behaviour of a characterisation
    """

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):
        BaseQueueEntry.__init__(self, view, data_model, view_set_queue_entry)
        self.edna_result = None
        self.auto_add_diff_plan = True

    def __getstate__(self):
        d = BaseQueueEntry.__getstate__(self)

        d["data_analysis_hwobj"] = (
            HWR.beamline.characterisation.name
            if HWR.beamline.characterisation
            else None
        )
        d["diffractometer_hwobj"] = (
            HWR.beamline.diffractometer.name if HWR.beamline.diffractometer else None
        )
        d["queue_model_hwobj"] = (
            HWR.beamline.queue_model.name if HWR.beamline.queue_model else None
        )
        d["session_hwobj"] = HWR.beamline.session.name if HWR.beamline.session else None

        return d

    def __setstate__(self, d):
        BaseQueueEntry.__setstate__(self, d)

    def execute(self):
        BaseQueueEntry.execute(self)

        if HWR.beamline.characterisation is not None:
            if self.get_data_model().wait_result:
                logging.getLogger("user_level_log").warning(
                    "Characterisation: Please wait ..."
                )
                self.start_char()
            else:
                logging.getLogger("user_level_log").warning(
                    "Characterisation: Please wait ..."
                )
                gevent.spawn(self.start_char)

    def start_char(self):
        char = self.get_data_model()
        characterisation_parameters = char.characterisation_parameters

        if characterisation_parameters.strategy_program != "None":
            log = logging.getLogger("user_level_log")
            self.get_view().setText(1, "Characterising")
            log.info("Characterising, please wait ...")
            reference_image_collection = char.reference_image_collection

            if HWR.beamline.characterisation is not None:
                edna_input = HWR.beamline.characterisation.input_from_params(
                    reference_image_collection, characterisation_parameters
                )

                self.edna_result = HWR.beamline.characterisation.characterise(
                    edna_input
                )

                if self.edna_result:
                    log.info("Characterisation completed.")

                    char.html_report = HWR.beamline.characterisation.get_html_report(
                        self.edna_result
                    )

                    try:
                        strategy_result = self.edna_result.getCharacterisationResult().getStrategyResult()
                    except Exception:
                        strategy_result = None

                    if strategy_result:
                        collection_plan = strategy_result.getCollectionPlan()
                    else:
                        collection_plan = None

                    if collection_plan:
                        if char.auto_add_diff_plan:
                            # default action
                            self.handle_diffraction_plan(self.edna_result, None)
                        else:
                            collections = HWR.beamline.characterisation.dc_from_output(
                                self.edna_result, char.reference_image_collection
                            )
                            char.diffraction_plan.append(collections)
                            HWR.beamline.queue_model.emit(
                                "diff_plan_available", (char, collections)
                            )

                        self.get_view().setText(1, "Done")
                    else:
                        self.get_view().setText(1, "No result")
                        self.status = QUEUE_ENTRY_STATUS.WARNING
                        log.warning(
                            "Characterisation completed "
                            + "successfully but without collection plan."
                        )
                else:
                    self.get_view().setText(1, "Charact. Failed")
                    log.error("EDNA-Characterisation completed with a failure.")

        char.set_executed(True)
        self.get_view().setHighlighted(True)

    def handle_diffraction_plan(self, edna_result, edna_collections):
        char = self.get_data_model()
        reference_image_collection = char.reference_image_collection

        dcg_model = char.get_parent()
        sample_data_model = dcg_model.get_parent()

        new_dcg_name = "Diffraction plan"
        new_dcg_num = dcg_model.get_parent().get_next_number_for_name(new_dcg_name)

        new_dcg_model = queue_model_objects.TaskGroup()
        new_dcg_model.set_enabled(False)
        new_dcg_model.set_name(new_dcg_name)
        new_dcg_model.set_number(new_dcg_num)
        new_dcg_model.set_origin(char._node_id)

        HWR.beamline.queue_model.add_child(sample_data_model, new_dcg_model)
        if edna_collections is None:
            edna_collections = HWR.beamline.characterisation.dc_from_output(
                edna_result, reference_image_collection
            )
        for edna_dc in edna_collections:
            path_template = edna_dc.acquisitions[0].path_template
            run_number = HWR.beamline.queue_model.get_next_run_number(path_template)
            path_template.run_number = run_number
            path_template.compression = char.diff_plan_compression

            edna_dc.set_enabled(char.run_diffraction_plan)
            edna_dc.set_name(path_template.get_prefix())
            edna_dc.set_number(path_template.run_number)
            HWR.beamline.queue_model.add_child(new_dcg_model, edna_dc)

        return edna_collections

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)
        self.get_view().setOn(True)
        self.get_view().setHighlighted(False)

    def post_execute(self):
        BaseQueueEntry.post_execute(self)

    def get_type_str(self):
        return "Characterisation"

    def stop(self):
        BaseQueueEntry.stop(self)
        HWR.beamline.characterisation.stop()
